import logging
import os
from datetime import timedelta
from enum import Enum
from importlib import import_module
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Union

from pydantic import BaseSettings, validator

from balsam import schemas
from balsam.util import validate_log_level


class LoginMethod(str, Enum):
    oauth_authcode = "oauth_authcode"
    oauth_device = "oauth_device"
    password = "password"


class OAuthProviderSettings(BaseSettings):
    class Config:
        env_prefix = "balsam_oauth_"
        case_sensitive = False
        env_file = ".env"
        extra = "forbid"

    # Environment must export BALSAM_OAUTH_CLIENT_ID, etc...
    client_id: str
    client_secret: str
    redirect_scheme: str = "https"
    redirect_path: str = "/auth/oauth/ALCF/callback"
    request_uri: str = "https://oauth2-dev.alcf.anl.gov/o/authorize/"
    token_uri: str = "https://oauth2-dev.alcf.anl.gov/o/token/"
    user_info_uri: str = "https://oauth2-dev.alcf.anl.gov/api/v1/user/"
    scope: str = "read_basic_user_data"
    device_code_lifetime: timedelta = timedelta(seconds=300)
    device_poll_interval: timedelta = timedelta(seconds=3)


class AuthSettings(BaseSettings):
    class Config:
        env_prefix = "balsam_auth_"
        case_sensitive = False
        env_file = ".env"
        extra = "forbid"

    secret_key = (
        "3f4abcaa16a006de8a7ef520a4c77c7d51b3ddb235c1ee2b99187f0f049eaa76"  # WARNING: Use real secret in production
    )
    algorithm = "HS256"
    token_ttl: timedelta = timedelta(hours=48)
    auth_method: str = "balsam.server.auth.user_from_token"
    login_methods: List[LoginMethod] = [LoginMethod.password]
    oauth_provider: Optional[OAuthProviderSettings] = None

    def get_auth_method(self) -> Callable[..., schemas.UserOut]:
        module, func = self.auth_method.rsplit(".", 1)
        return getattr(import_module(module), func)  # type: ignore

    @validator("oauth_provider", always=True)
    def get_oauth_config(cls, v: Optional[OAuthProviderSettings]) -> Optional[OAuthProviderSettings]:
        if v is None and os.environ.get("BALSAM_OAUTH_CLIENT_ID"):
            return OAuthProviderSettings()
        return v


class Settings(BaseSettings):
    class Config:
        env_prefix = "balsam_"
        case_sensitive = False
        env_file = ".env"
        extra = "forbid"

    database_url: str = "postgresql://postgres@localhost:5432/balsam"
    auth: AuthSettings = AuthSettings()
    redis_params: Dict[str, Any] = {"unix_socket_path": "/tmp/redis-balsam.server.sock"}
    log_level: Union[str, int] = logging.INFO
    log_dir: Optional[Path]

    @validator("log_level", always=True)
    def validate_balsam_log_level(cls, v: Union[str, int]) -> int:
        return validate_log_level(v)

    @validator("log_dir", always=True)
    def validate_log_dir(cls, v: Optional[Path]) -> Optional[Path]:
        if v is None:
            return None
        if not v.exists():
            v.mkdir(parents=True, exist_ok=False)
        elif not v.is_dir():
            raise ValueError(f"{v} is not a valid log directory")
        v.chmod(0o700)
        return v

    def serialize_without_secrets(self) -> str:
        return self.json(
            exclude={
                "auth": {
                    "secret_key": ...,
                    "oauth_provider": {
                        "client_secret": ...,
                    },
                }
            }
        )



from .job_source import FixedDepthJobSource, SynchronousJobSource
from .script_template import ScriptTemplate
from .status_updater import BulkStatusUpdater

__all__ = [
    "FixedDepthJobSource",
    "SynchronousJobSource",
    "BulkStatusUpdater",
    "ScriptTemplate",
]



"""
multiprocessing.Process has a hard-coded excepthook that prints exceptions to stderr
We use this to ensure that all exceptions get logged uniformly before bubbling up to stderr
"""
import multiprocessing
import sys

from .log import log_uncaught_exceptions


class Process(multiprocessing.Process):
    def _run(self) -> None:
        raise NotImplementedError

    def run(self) -> None:
        try:
            self._run()
        except NotImplementedError:
            try:
                super().run()
            except Exception:
                log_uncaught_exceptions(*sys.exc_info())
                raise
        except Exception:
            log_uncaught_exceptions(*sys.exc_info())
            raise



import inspect
import logging
import os
import re
import shlex
import sys
from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Set, Tuple, Type, Union

import jinja2
import jinja2.meta

from balsam.schemas import DeserializeError, JobState, SerializeError, deserialize, get_source, serialize

if TYPE_CHECKING:
    from balsam.client import RESTClient

    from .models import App, Job, Site

PARAM_PATTERN = re.compile(r"{{(.*?)}}")
logger = logging.getLogger(__name__)


def is_appdef(attr: type) -> bool:
    return (
        isinstance(attr, type)
        and issubclass(attr, ApplicationDefinition)
        and attr.__name__ != "ApplicationDefinition"
    )


def is_valid_template(command_template: str) -> bool:
    for param in re.findall(PARAM_PATTERN, command_template):
        if not param.strip().isidentifier():
            return False
    return True


def validate_dict_parameter(param_name: str, param: Dict[str, str]) -> None:
    if "required" not in param:
        raise AttributeError(f"ApplicationDefinition parameter {param_name} must set 'required' key as True/False")
    if not param["required"] and param.get("default") is None:
        raise AttributeError(f"ApplicationDefinition optional parameter {param_name} must set a 'default' str value")


def detect_template_parameters(command_template: str) -> Set[str]:
    ctx = jinja2.Environment().parse(command_template)
    detected_params: Set[str] = jinja2.meta.find_undeclared_variables(ctx)
    return detected_params


def check_extraneous_parameters(detected_params: Set[str], dict_params: Set[str]) -> None:
    extraneous = dict_params.difference(detected_params)
    if extraneous:
        raise AttributeError(
            f"App has extraneous keys "
            f"in the `parameters` dictionary: {extraneous}"
            f"\nparameters dict: {dict_params}"
            f"\nDetected parameters in command_template: {detected_params}"
        )


def chunk_str(s: str, chunksize: int) -> List[str]:
    num_chunks = len(s) // chunksize
    chunks = [s[i * chunksize : (i + 1) * chunksize] for i in range(num_chunks + 1)]
    chunks = [shlex.quote(chunk) for chunk in chunks]
    return chunks


def metadata_from_signature(param: inspect.Parameter) -> Dict[str, Any]:
    if param.annotation:
        if hasattr(param.annotation, "__name__"):
            help_text = param.annotation.__name__
        else:
            help_text = str(param.annotation)
    else:
        help_text = ""

    empty = inspect.Parameter.empty
    if param.default is empty:
        return {
            "required": True,
            "default": None,
            "help": help_text,
        }
    else:
        return {
            "required": False,
            "default": param.default,
            "help": help_text,
        }


class AppType(str, Enum):
    SHELL_CMD = "SHELL_CMD"
    PY_FUNC = "PY_FUNC"


class ApplicationDefinitionMeta(type):
    environment_variables: Dict[str, str]
    command_template: str
    parameters: Dict[str, Any]
    transfers: Dict[str, Any]
    cleanup_files: List[str]
    _default_params: Dict[str, str]
    _app_type: AppType
    run: Optional[Callable[..., Any]]
    _client: Optional["RESTClient"]

    @property
    def source_code(cls) -> str:
        return get_source(cls)

    @property
    def _Site(cls) -> Type["Site"]:
        if cls._client is None:
            raise AttributeError(f"Client has not been set: call {cls.__name__}._set_client prior to this method")
        return cls._client.Site

    @property
    def _App(cls) -> Type["App"]:
        if cls._client is None:
            raise AttributeError(f"Client has not been set: call {cls.__name__}._set_client prior to this method")
        return cls._client.App

    def _setup_shell_app(cls) -> "ApplicationDefinitionMeta":
        cls._app_type = AppType.SHELL_CMD
        cls.command_template = cls.command_template.strip()
        if not is_valid_template(cls.command_template):
            raise AttributeError(
                f"ApplicationDefinition {cls.__name__} command_template parameters must be valid Python identifiers"
            )

        for param_name, param in cls.parameters.items():
            validate_dict_parameter(param_name, param)

        detected_params = detect_template_parameters(cls.command_template)
        dict_params = set(cls.parameters.keys())
        check_extraneous_parameters(detected_params, dict_params)

        for param in detected_params.difference(dict_params):
            cls.parameters[param] = {
                "required": True,
                "default": None,
                "help": "",
            }
        cls._default_params = {k: v["default"] for k, v in cls.parameters.items() if not v["required"]}
        return cls

    def _setup_py_app(cls) -> "ApplicationDefinitionMeta":
        assert callable(cls.run)
        cls._app_type = AppType.PY_FUNC

        signature = inspect.signature(cls.run)
        if "self" not in signature.parameters:
            raise TypeError(f"ApplicationDefinition {cls.__name__}.run{signature} is missing the `self` argument")

        cls.parameters = {}
        for param_name, param in signature.parameters.items():
            if param_name == "self":
                continue
            cls.parameters[param_name] = metadata_from_signature(param)
        return cls

    def __new__(mcls, name: str, bases: Tuple[Any, ...], attrs: Dict[str, Any]) -> "ApplicationDefinitionMeta":
        attrs["parameters"] = attrs.pop("parameters", {}).copy()
        cls = super().__new__(mcls, name, bases, attrs)
        if not bases:
            return cls

        if "site" not in attrs:
            raise AttributeError(
                f"ApplicationDefinition {name} must contain the `site` attribute, set to a site id, name, or Site object"
            )

        has_command_template = isinstance(getattr(cls, "command_template", None), str)
        has_run_function = callable(getattr(cls, "run", None))

        if has_command_template and has_run_function:
            raise AttributeError(
                f"ApplicationDefinition {name} must not define both `run` function and `command_template`"
            )
        elif has_command_template:
            return mcls._setup_shell_app(cls)
        elif has_run_function:
            return mcls._setup_py_app(cls)
        else:
            raise AttributeError(
                f"ApplicationDefinition {name} must define either a `run` function or a `command_template`"
            )


AppDefType = Type["ApplicationDefinition"]


class ApplicationDefinition(metaclass=ApplicationDefinitionMeta):
    class ModuleLoadError(Exception):
        pass

    ARG_CHUNK_SIZE: int = 128_000
    environment_variables: Dict[str, str] = {}
    command_template: str
    parameters: Dict[str, Any] = {}
    transfers: Dict[str, Any] = {}
    cleanup_files: List[str] = []
    _default_params: Dict[str, str]
    run: Optional[Callable[..., Any]] = None
    site: Union[int, str, "Site"]
    python_exe: str = sys.executable
    _site_id: Optional[int] = None
    _client: Optional["RESTClient"] = None
    _app_type: AppType
    __app_id__: Optional[int] = None
    _app_name_cache: Dict[Tuple[Optional[str], str], AppDefType] = {}
    _app_id_cache: Dict[int, AppDefType] = {}
    _serialized_class: Optional[str] = None

    @staticmethod
    def _set_client(client: "RESTClient") -> None:
        ApplicationDefinition._client = client

    def __init__(self, job: "Job") -> None:
        self.job = job

    def get_arg_str(self) -> str:
        if self._app_type == AppType.SHELL_CMD:
            return self._render_shell_command()
        elif self._app_type == AppType.PY_FUNC:
            return self._render_pyrunner_command()
        else:
            raise TypeError(f"Invalid _app_type: {self._app_type}")

    def get_environ_vars(self) -> Dict[str, str]:
        envs = os.environ.copy()
        envs.update(self.environment_variables)
        envs["BALSAM_JOB_ID"] = str(self.job.id)
        if self.job.threads_per_rank > 1:
            envs["OMP_NUM_THREADS"] = str(self.job.threads_per_rank)
        return envs

    def _render_pyrunner_command(self) -> str:
        assert self.__app_id__ is not None
        assert self.job._read_model is not None
        assert self._serialized_class is not None
        app_id = int(self.__app_id__)

        app_payload = self._serialized_class
        app_chunks = chunk_str(app_payload, self.ARG_CHUNK_SIZE)
        num_app_chunks = len(app_chunks)

        job_payload = self.job._read_model.json()
        job_chunks = chunk_str(job_payload, self.ARG_CHUNK_SIZE)

        args = f"{app_id} {num_app_chunks} {' '.join(app_chunks)} {' '.join(job_chunks)}"
        return f"{self.python_exe} -m balsam.site.launcher.python_runner {args}"

    def _render_shell_command(self) -> str:
        """
        Args:
            - arg_dict: value for each required parameter
        Returns:
            - str: shell command with safely-escaped parameters
            Use shlex.split() to split the result into args list.
            Do *NOT* use string.join on the split list: unsafe!
        """
        arg_dict = {**self._default_params, **self.job.get_parameters()}
        diff = set(self.parameters.keys()).difference(arg_dict.keys())
        if diff:
            raise ValueError(f"Missing required args: {diff} (only got: {arg_dict})")

        sanitized_args = {key: shlex.quote(str(arg_dict[key])) for key in self.parameters}
        return jinja2.Template(self.command_template).render(sanitized_args)

    def preprocess(self) -> None:
        self.job.state
        self.job.state = JobState.preprocessed

    def postprocess(self) -> None:
        self.job.state = JobState.postprocessed

    def shell_preamble(self) -> Union[str, List[str]]:
        return []

    def handle_timeout(self) -> None:
        self.job.state = JobState.restart_ready

    def handle_error(self) -> None:
        self.job.state = JobState.failed

    @classmethod
    def resolve_site_id(cls) -> int:
        from balsam._api.models import Site

        if cls._site_id is not None:
            return cls._site_id
        elif isinstance(cls.site, str):
            site = cls._Site.objects.get(name=cls.site)
            assert site.id is not None
            cls._site_id = site.id
        elif isinstance(cls.site, Site):
            if cls.site.id is None:
                raise ValueError(f"{cls.__name__}.site does not have an ID set: {cls.site}")
            cls._site_id = cls.site.id
        else:
            if not isinstance(cls.site, int):
                raise ValueError(
                    f"{cls.__name__}.site must be a string, integer, or Site object. Got: {type(cls.site)}"
                )
            cls._site_id = cls.site
        return cls._site_id

    @staticmethod
    def load_by_site(site: Union[int, str, "Site"]) -> Dict[str, AppDefType]:
        AppModel: "Type[App]" = ApplicationDefinition._App

        lookup: Dict[str, Union[str, int]]
        if isinstance(site, int):
            lookup = {"site_id": site}
        elif isinstance(site, str):
            lookup = {"site_name": site}
        elif isinstance(site, ApplicationDefinition._Site):
            assert site.id is not None
            lookup = {"site_id": site.id}
        else:
            raise ValueError("site must be an int, str, or Site object.")

        api_apps = AppModel.objects.filter(**lookup)  # type: ignore
        apps_by_name = {}
        for app in api_apps:
            apps_by_name[app.name] = ApplicationDefinition.from_serialized(app)
            assert app.id is not None
            ApplicationDefinition._app_id_cache[app.id] = apps_by_name[app.name]
        return apps_by_name

    @classmethod
    def load_by_name(cls, app_name: str, site_name: Optional[str] = None) -> AppDefType:
        app_key = (site_name, app_name)
        if app_key not in cls._app_name_cache:
            logger.debug(f"App Cache miss: fetching app {app_key}")
            app: "App" = cls._App.objects.get(site_name=site_name, name=app_name)
            assert app.id is not None
            app_def = cls.from_serialized(app)
            cls._app_name_cache[app_key] = app_def
            cls._app_id_cache[app.id] = app_def
        return cls._app_name_cache[app_key]

    @classmethod
    def load_by_id(cls, app_id: int) -> AppDefType:
        if app_id not in cls._app_id_cache:
            logger.debug(f"App Cache miss: fetching app {app_id}")
            api_app: "App" = cls._App.objects.get(id=app_id)
            app_def = cls.from_serialized(api_app)
            cls._app_id_cache[app_id] = app_def
        return cls._app_id_cache[app_id]

    @classmethod
    def sync(cls, rename_from: Optional[str] = None) -> None:
        app_dict = cls.to_dict()
        existing_app: Optional[App] = None
        AppModel: Type["App"] = cls._App
        if rename_from:
            existing_app = AppModel.objects.get(site_id=app_dict["site_id"], name=rename_from)
            logger.info(f"Renaming App(id={existing_app.id}): {rename_from} -> {app_dict['name']}")
        else:
            try:
                existing_app = AppModel.objects.get(site_id=app_dict["site_id"], name=app_dict["name"])
                logger.info(f"Updating App(id={existing_app.id}, name={app_dict['name']})")
            except AppModel.DoesNotExist:
                pass

        if existing_app:
            for k, v in app_dict.items():
                setattr(existing_app, k, v)
            existing_app.save()
            cls.__app_id__ = existing_app.id
            assert existing_app.id is not None
            cls._app_id_cache[existing_app.id] = cls
        else:
            new_app = AppModel.objects.create(**app_dict)
            logger.info(f"Created new App(id={new_app.id}, name={app_dict['name']})")
            cls.__app_id__ = new_app.id
            assert new_app.id is not None
            cls._app_id_cache[new_app.id] = cls

    @classmethod
    def submit(
        cls,
        workdir: Path,
        tags: Optional[Dict[str, str]] = None,
        data: Optional[Dict[str, Any]] = None,
        return_code: Optional[int] = None,
        num_nodes: int = 1,
        ranks_per_node: int = 1,
        threads_per_rank: int = 1,
        threads_per_core: int = 1,
        launch_params: Optional[Dict[str, str]] = None,
        gpus_per_rank: float = 0,
        node_packing_count: int = 1,
        wall_time_min: int = 0,
        parent_ids: Set[int] = set(),
        transfers: Optional[Dict[str, str]] = None,
        save: bool = True,
        **app_params: Any,
    ) -> "Job":
        """
        Construct a new Job object.  If save=True (default), the Job
        will be saved to the API automatically.  Use save=False to create
        in-memory Jobs only, which can then be created in bulk by passing
        into Job.objects.bulk_create().

        workdir:            Job path relative to site data/ folder.
        tags:               Custom key:value string tags.
        data:               Arbitrary JSON-able data dictionary.
        return_code:        Return code from last execution of this Job.
        num_nodes:          Number of compute nodes needed.
        ranks_per_node:     Number of MPI processes per node.
        threads_per_rank:   Logical threads per process.
        threads_per_core:   Logical threads per CPU core.
        launch_params:      Optional pass-through parameters to MPI application launcher.
        gpus_per_rank:      Number of GPUs per process.
        node_packing_count: Maximum number of concurrent runs per node.
        wall_time_min:      Optional estimate of Job runtime. All else being equal, longer Jobs tend to run first.
        parent_ids:         Set of parent Job IDs (dependencies).
        transfers:          TransferItem dictionary. One key:JobTransferItem pair for each slot defined on the App.
        save:               Whether to save the Job to the API immediately.
        **app_params:       Parameters passed to App at runtime.
        """
        if cls.__app_id__ is None:
            cls.sync()
        assert cls.__app_id__ is not None
        assert cls._client is not None
        job_kwargs = {k: v for k, v in locals().items() if k not in ["cls", "__class__"] and v is not None}
        job_kwargs["parameters"] = app_params
        job_kwargs["app_id"] = cls.__app_id__
        job = cls._client.Job(**job_kwargs)
        if save:
            job.save()
        return job

    @classmethod
    def to_dict(cls) -> Dict[str, Any]:
        try:
            serialized_class = serialize(cls)
        except SerializeError as exc:
            logger.error(f"Please fix {cls.__name__}: can't serialize class due to: {exc}")
            raise
        cls._serialized_class = serialized_class
        return dict(
            site_id=cls.resolve_site_id(),
            name=cls.__name__,
            serialized_class=serialized_class,
            source_code=cls.source_code,
            description=(cls.__doc__ or "").strip(),
            parameters=cls.parameters,
            transfers=cls.transfers,
        )

    @staticmethod
    def from_serialized(app: "App") -> AppDefType:
        if app.id is None:
            raise ValueError("Cannot deserialize App with id=None")

        try:
            cls: AppDefType = deserialize(app.serialized_class)
        except DeserializeError as exc:
            raise DeserializeError(f"Failed to deserialize App(id={app.id}, name={app.name}): {exc}") from exc
        if not is_appdef(cls):
            raise TypeError(f"Deserialized {cls.__name__} is not an ApplicationDefinition subclass")

        cls.__app_id__ = app.id
        cls._serialized_class = app.serialized_class
        return cls



from datetime import datetime
from typing import Dict, List, Optional, Set

from pydantic import BaseModel, Field

from .job import RUNNABLE_STATES, JobState


class SessionCreate(BaseModel):
    site_id: int = Field(..., description="Site id of the running Session")
    batch_job_id: Optional[int] = Field(None, description="Associated batchjob id")


class SessionOut(BaseModel):
    id: int = Field(..., description="Session id")
    site_id: int = Field(..., description="Site id of the running Session")
    batch_job_id: Optional[int] = Field(None, description="Associated batchjob id")
    heartbeat: datetime = Field(..., description="Last heartbeat received from Session")

    class Config:
        orm_mode = True


class SessionAcquire(BaseModel):
    max_num_jobs: int
    max_wall_time_min: Optional[int]
    max_nodes_per_job: Optional[int]
    max_aggregate_nodes: Optional[float]
    serial_only: bool = False
    filter_tags: Dict[str, str]
    states: Set[JobState] = RUNNABLE_STATES
    app_ids: Set[int] = set()


class PaginatedSessionsOut(BaseModel):
    count: int
    results: List[SessionOut]



import logging
import multiprocessing
import os
import socket
import time
from pathlib import Path
from typing import Any

from balsam.config import SiteConfig
from balsam.util import SigHandler

logger = logging.getLogger("balsam.site.service.main")


class PIDFile:
    def __init__(self, site_path: Path) -> None:
        self.path: Path = site_path.joinpath("balsam-service.pid")

    def __enter__(self) -> None:
        if self.path.exists():
            raise RuntimeError(
                f"{self.path} already exists: will not start Balsam service because it's already running. "
                f"To restart, first stop the Balsam service with `balsam service stop`"
            )
        with open(self.path, "w") as fp:
            fp.write(f"{socket.gethostname()}\n{os.getpid()}")

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        self.path.unlink()


def main(config: SiteConfig, run_time_sec: int) -> None:
    start_time = time.time()
    sig_handler = SigHandler()
    config.enable_logging(basename="service")
    m, s = divmod(run_time_sec, 60)
    h, m = divmod(m, 60)
    logger.info(f"Launching service for {h:02d}h:{m:02d}m:{s:02d}s")

    config.update_site_from_config()

    services = config.build_services()

    for service in services:
        attrs = {k: v for k, v in service.__dict__.items() if not k.startswith("_")}
        logger.info(f"Starting service: {service.__class__.__name__}")
        logger.info(str(attrs))
        service.start()

    while not sig_handler.wait_until_exit(timeout=60):
        elapsed = time.time() - start_time
        if elapsed > run_time_sec:
            logger.info("Exceeded max service runtime")
            break
        remaining = int(run_time_sec - elapsed)
        m, s = divmod(remaining, 60)
        if (m, s) >= (60, 0):
            h, m = divmod(m, 60)
            logger.info(f"{h:02d}h:{m:02d}m remaining")
        else:
            logger.info(f"{m:02d}m:{s:02d}s remaining")

    logger.info("Terminating services...")
    for service in services:
        service.terminate()

    logger.info("Waiting for service processes to join")
    for service in services:
        service.join()

    logger.info("Balsam service: exit main loop")


if __name__ == "__main__":
    multiprocessing.set_start_method("fork", force=True)
    config = SiteConfig()
    run_time_sec = int(config.client.expires_in.total_seconds() - 120)
    if run_time_sec < 10:
        raise RuntimeError(
            f"Client authorization will expire in {run_time_sec+120} seconds. Please refresh auth with `balsam login`."
        )

    with PIDFile(config.site_path):
        main(config, run_time_sec)
    logger.info(f"Balsam service [pid {os.getpid()}] goodbye: exited PIDFile context manager.")



from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set, Union, cast

import click
import yaml

from balsam.schemas import JobState, JobTransferItem

from .utils import filter_by_sites, load_client, table_print, validate_tags

if TYPE_CHECKING:
    from balsam._api.models import App, AppQuery, JobQuery
    from balsam.client import RESTClient  # noqa: F401


@click.group()
def job() -> None:
    """
    Create and monitor Balsam Jobs
    """
    pass


def validate_state(ctx: Any, param: Any, value: Union[None, str, JobState]) -> Union[None, JobState]:
    if value is None:
        return value
    if not JobState.is_valid(value):
        raise click.BadParameter(f"Invalid state {value}")
    return cast(JobState, value)


def fetch_app(app_qs: "AppQuery", app_str: str) -> "App":
    App = app_qs._manager._model_class
    lookup: Dict[str, Any]
    if app_str.isdigit():
        lookup = {"id": int(app_str)}
    else:
        lookup = {"name": app_str}
    try:
        app = app_qs.get(**lookup)
    except App.DoesNotExist:
        raise click.BadParameter(f"No App matching criteria {lookup}")
    return app


def validate_set(all_params: Set[str], required_params: Set[str], provided_params: Set[str]) -> None:
    missing_params = required_params.difference(provided_params)
    extraneous_params = provided_params.difference(all_params)
    if missing_params:
        raise click.BadParameter(f"Missing required parameters (-p): {missing_params}")
    if extraneous_params:
        raise click.BadParameter(f"Extraneous parameters (-p): {extraneous_params}")


def validate_parameters(parameters: List[str], app: "App") -> Dict[str, str]:
    params = validate_tags(None, None, parameters)
    all_params = set(app.parameters.keys())
    required_params = {k for k in all_params if app.parameters[k].required}
    provided = set(params.keys())
    validate_set(all_params, required_params, provided)
    return params


def validate_transfers(transfer_args: List[str], app: "App") -> Dict[str, JobTransferItem]:
    transfers = validate_tags(None, None, transfer_args)
    all_transfers = set(app.transfers.keys())
    required_transfers = {k for k in all_transfers if app.transfers[k].required}
    provided = set(transfers.keys())
    validate_set(all_transfers, required_transfers, provided)

    transfers_by_name: Dict[str, JobTransferItem] = {}
    for name in transfers:
        try:
            loc, path = transfers[name].split(":")
        except ValueError:
            raise click.BadParameter("Transfers must take the form LOCATION_ALIAS:PATH")
        transfers_by_name[name] = JobTransferItem(location_alias=loc, path=path)
    return transfers_by_name


def validate_parents(parent_ids: List[int], client: "RESTClient") -> None:
    if not parent_ids:
        return None
    jobs = list(client.Job.objects.filter(id=parent_ids))
    if len(jobs) < len(parent_ids):
        job_ids = [j.id for j in jobs]
        missing_ids = [i for i in parent_ids if i not in job_ids]
        raise click.BadParameter(f"Could not find parent job ids {missing_ids}")
    return None


@job.command()
@click.option("-w", "--workdir", required=True, type=str, help="Job directory (relative to data/)")
@click.option("-a", "--app", "app_str", required=True, type=str, help="App ID or name (module.ClassName)")
@click.option(
    "-tag", "--tag", "tags", multiple=True, type=str, callback=validate_tags, help="Job tags (--tag KEY=VALUE)"
)
@click.option(
    "-p",
    "--param",
    "parameters",
    multiple=True,
    type=str,
    help="App command template parameters (--param name=value)",
)
@click.option("-n", "--num-nodes", default=1, type=int, help="Number of compute nodes to run on", show_default=True)
@click.option("-rpn", "--ranks-per-node", default=1, type=int, help="MPI ranks per node", show_default=True)
@click.option("-tpr", "--threads-per-rank", default=1, type=int, help="Threads per process/rank", show_default=True)
@click.option("-tpc", "--threads-per-core", default=1, type=int, help="Threads per CPU core", show_default=True)
@click.option("-g", "--gpus-per-rank", default=0, type=float, help="GPUs per process/rank", show_default=True)
@click.option(
    "-npc", "--node-packing-count", default=1, type=int, help="Max concurrent runs per node", show_default=True
)
@click.option(
    "-lp",
    "--launch-param",
    "launch_params",
    multiple=True,
    type=str,
    callback=validate_tags,
    help="Pass-through parameters to MPI launcher (-lp name=value)",
)
@click.option("-t", "--wall-time-min", default=1, type=int)
@click.option(
    "-pid",
    "--parent-id",
    "parent_ids",
    multiple=True,
    type=int,
    help="Job dependencies given as one or many parent IDs",
    show_default=True,
)
@click.option(
    "-s",
    "--stage-data",
    "transfer_args",
    multiple=True,
    type=str,
    help="Transfer slots given as TRANSFER_SLOT=LOCATION_ALIAS:/path/to/file",
)
@click.option("-y", "--yes", "force_create", is_flag=True, default=False)
@click.option("--site", "site_selector", default="", help="Site ID or path fragment")
def create(
    workdir: str,
    app_str: str,
    tags: Dict[str, str],
    parameters: List[str],
    num_nodes: int,
    ranks_per_node: int,
    threads_per_rank: int,
    threads_per_core: int,
    gpus_per_rank: int,
    node_packing_count: int,
    launch_params: Dict[str, str],
    wall_time_min: int,
    parent_ids: List[int],
    transfer_args: List[str],
    site_selector: str,
    force_create: bool,
) -> None:
    """
    Add a new Balsam Job

    Examples:

    Create a Job in workdir "data/test/1" running app `demo.Hello` with parameter name="world!"

        balsam job create -w test/1 -a demo.Hello -p name="world!"
    """
    client: RESTClient = load_client()
    if Path(workdir).is_absolute():
        raise click.BadParameter("workdir must be a relative path: cannot start with '/'")

    app_qs = filter_by_sites(client.App.objects.all(), site_selector)
    app = fetch_app(app_qs, app_str)
    assert app.id is not None, "Could not resolve application ID"
    parameters_dict = validate_parameters(parameters, app)
    transfers = validate_transfers(transfer_args, app)
    validate_parents(parent_ids, client)

    job = client.Job(
        workdir=Path(workdir),
        app_id=app.id,
        tags=tags,
        parameters=parameters_dict,
        num_nodes=num_nodes,
        ranks_per_node=ranks_per_node,
        threads_per_rank=threads_per_rank,
        threads_per_core=threads_per_core,
        gpus_per_rank=gpus_per_rank,
        node_packing_count=node_packing_count,
        launch_params=launch_params,
        wall_time_min=wall_time_min,
        parent_ids=set(parent_ids),
        transfers=transfers,
    )
    click.echo(yaml.dump(job.display_dict(), sort_keys=False, indent=4))
    if force_create or click.confirm("Do you want to create this Job?"):
        job.save()
        click.echo(f"Added Job id={job.id}")


def count_by_state(job_qs: "JobQuery", verbose: bool) -> None:
    state_data: List[Dict[str, Any]] = []
    # job_qs
    for state in JobState:
        state_count = job_qs.filter(state=state).count()
        assert state_count is not None
        if state_count > 0 or verbose:
            state_dict = {"State": state.value, "Count": state_count}
            state_data.append(state_dict)

    table_print(state_data)


def list_verbose(job_qs: "JobQuery") -> None:
    for job in job_qs:
        click.echo(yaml.dump(job.display_dict(), sort_keys=False, indent=4))
        click.echo("---\n")


def list_table(job_qs: "JobQuery", client: "RESTClient") -> None:
    sites = {s.id: s for s in client.Site.objects.all()}
    apps = {a.id: a for a in client.App.objects.all()}
    data = []
    for j in job_qs:
        app = apps[j.app_id]
        site = sites[app.site_id]
        assert j.state is not None
        jdict = {
            "ID": j.id,
            "Site": f"{site.name}",
            "App": app.name,
            "Workdir": j.workdir.as_posix(),
            "State": j.state.value,
            "Tags": j.tags,
        }
        data.append(jdict)
    table_print(data)


@job.command()
@click.option("-t", "--tag", "tags", multiple=True, type=str, callback=validate_tags)
@click.option("-s", "--state", type=str, callback=validate_state)
@click.option("-ns", "--exclude-state", type=str, callback=validate_state)
@click.option("--id", type=str)
@click.option("--by-state", type=bool, default=False, is_flag=True)
@click.option("-w", "--workdir", type=str)
@click.option("--site", "site_selector", default="")
@click.option("-v", "--verbose", is_flag=True)
def ls(
    tags: List[str],
    state: Optional[JobState],
    exclude_state: Optional[JobState],
    id: Optional[int],
    by_state: Optional[bool],
    workdir: Optional[str],
    verbose: bool,
    site_selector: str,
) -> None:
    """
    List Balsam Jobs

    1) Filter by Site ID or Path fragments (comma-separated)

        balsam job ls --site=123,my-cori-site

    2) Select Jobs by their tags

        balsam job ls --tag experiment=XPCS --tag system=H2O

    3) Select Jobs by their state

        balsam job ls --state JOB_FINISHED --tag system=H2O

    4) Summarize Jobs by their state

        balsam job ls --by-state [--tag system=H20]

    5) Select a specific job by ID

        balsam job ls --id [id]
    """
    client = load_client()
    job_qs = filter_by_sites(client.Job.objects.all(), site_selector)
    if tags:
        job_qs = job_qs.filter(tags=tags)
    if state:
        job_qs = job_qs.filter(state=state)
    if exclude_state:
        job_qs = job_qs.filter(state__ne=exclude_state)
    if workdir:
        job_qs = job_qs.filter(workdir__contains=workdir)
    if id:
        job_qs = job_qs.filter(id=[id])

    if by_state:
        count_by_state(job_qs, verbose)
    elif verbose:
        list_verbose(job_qs)
    else:
        list_table(job_qs, client)


@job.command()
@click.option("-i", "--id", "job_ids", multiple=True, type=int)
@click.option("-t", "--tag", "tags", multiple=True, type=str, callback=validate_tags)
@click.option("--all", is_flag=True, default=False)
def rm(job_ids: List[int], tags: List[str], all: bool) -> None:
    """
    Remove Jobs

    1) Remove Jobs by ID

        balsam job rm --id 41 --id 42 --id 43

    2) Remove Jobs by Tags

        balsam job rm --tag workflow=temp-test

    3) Remove all jobs (DANGER!)

        balsam job rm --all
    """
    client: RESTClient = load_client()
    jobs = client.Job.objects.all()
    if job_ids:
        jobs = jobs.filter(id=job_ids)
    elif tags:
        jobs = jobs.filter(tags=tags)
    elif all:
        click.echo("THIS WILL DELETE ALL JOBS! CAUTION!")
        pass
    else:
        raise click.BadParameter("Provide either list of Job ids or tags to delete")
    count = jobs.count()
    assert count is not None
    if count < 1:
        click.echo("No jobs match deletion query")
    elif click.confirm(f"Really delete {count} jobs?"):
        jobs.delete()



import logging
import multiprocessing

import click

from balsam.site.launcher._serial_mode_master import master_main
from balsam.site.launcher._serial_mode_worker import worker_main

logger = logging.getLogger("balsam.site.launcher")


@click.command()
@click.option("--wall-time-min", type=int)
@click.option("--master-address")
@click.option("--run-master", is_flag=True, default=False)
@click.option("--log-filename")
@click.option("--num-workers", type=int)
@click.option("--filter-tags")
def entry_point(
    wall_time_min: int, master_address: str, run_master: bool, log_filename: str, num_workers: int, filter_tags: str
) -> None:
    master_host, master_port = master_address.split(":")
    if run_master:
        logger.info("Running serial mode master at", master_address)
        master_main(wall_time_min, int(master_port), log_filename, num_workers, filter_tags)
    else:
        logger.info("Running serial mode worker")
        worker_main(master_host, int(master_port), log_filename)


if __name__ == "__main__":
    multiprocessing.set_start_method("fork", force=True)
    entry_point()



import abc
from collections import defaultdict
from enum import Enum
from pathlib import Path
from typing import Any, Dict, List, Sequence, Tuple, Type

from pydantic import BaseModel


class TransferSubmitError(Exception):
    pass


class TransferRetryableError(Exception):
    pass


class TaskState(str, Enum):
    active = "active"
    inactive = "inactive"
    done = "done"
    error = "error"


class TaskInfo(BaseModel):
    task_id: str
    state: TaskState
    info: Dict[str, Any]


class TransferInterface(abc.ABC):
    _registry: Dict[str, Type["TransferInterface"]] = {}

    @abc.abstractmethod
    def submit_task(
        self,
        remote_loc: str,
        direction: str,
        transfer_paths: List[Tuple[Path, Path, bool]],
    ) -> str:
        raise NotImplementedError

    @staticmethod
    @abc.abstractmethod
    def _poll_tasks(task_ids: Sequence[str]) -> List[TaskInfo]:
        raise NotImplementedError

    @staticmethod
    def poll_tasks(task_ids: List[str]) -> List[TaskInfo]:
        ids_by_protocol = defaultdict(list)
        for task_id in task_ids:
            protocol, id = str(task_id).split(":")
            ids_by_protocol[protocol].append(id)

        result = []
        for protocol, id_list in ids_by_protocol.items():
            cls = TransferInterface._registry[protocol]
            task_infos = cls._poll_tasks(id_list)
            result.extend(task_infos)
        return result



from .cobalt_sched import CobaltScheduler
from .local import LocalProcessScheduler
from .lsf_sched import LsfScheduler
from .scheduler import (
    SchedulerDeleteError,
    SchedulerError,
    SchedulerInterface,
    SchedulerNonZeroReturnCode,
    SchedulerSubmitError,
)
from .slurm_sched import SlurmScheduler

__all__ = [
    "LocalProcessScheduler",
    "SchedulerInterface",
    "CobaltScheduler",
    "SlurmScheduler",
    "LsfScheduler",
    "SchedulerError",
    "SchedulerSubmitError",
    "SchedulerDeleteError",
    "SchedulerNonZeroReturnCode",
]



from datetime import datetime
from enum import Enum
from typing import Dict, List, Optional

from pydantic import BaseModel, Field


class JobMode(str, Enum):
    serial = "serial"
    mpi = "mpi"


class BatchJobOrdering(str, Enum):
    start_time = "start_time"
    start_time_desc = "-start_time"


class BatchJobState(str, Enum):
    pending_submission = "pending_submission"
    queued = "queued"
    running = "running"
    finished = "finished"
    submit_failed = "submit_failed"
    pending_deletion = "pending_deletion"


class SchedulerJobStatus(BaseModel):
    scheduler_id: int
    state: BatchJobState
    queue: str
    num_nodes: int
    wall_time_min: int
    project: str
    time_remaining_min: int
    queued_time_min: int


class SchedulerBackfillWindow(BaseModel):
    num_nodes: int
    wall_time_min: int


class SchedulerJobLog(BaseModel):
    start_time: Optional[datetime]
    end_time: Optional[datetime]


class BatchJobPartition(BaseModel):
    job_mode: JobMode = Field(..., example="mpi")
    num_nodes: int = Field(..., example=128)
    filter_tags: Dict[str, str] = Field({}, example={"system": "H2O", "calc_type": "energy"})


class BatchJobBase(BaseModel):
    num_nodes: int = Field(..., example=128, description="Requested number of nodes for this allocation")
    wall_time_min: int = Field(..., example=60, description="Requested wall clock time for this allocation")
    job_mode: JobMode = Field(..., example="mpi", description="Balsam launcher execution mode (if single partition)")
    optional_params: Dict[str, str] = Field(
        {},
        example={"require_ssds": "1"},
        description="Optional pass-through parameters submitted with the batchjob script",
    )
    filter_tags: Dict[str, str] = Field(
        {}, example={"system": "H2O", "calc_type": "energy"}, description="Only run Jobs containing these tags"
    )
    partitions: Optional[List[BatchJobPartition]] = Field(
        None,
        example=[
            {"job_mode": "mpi", "num_nodes": 1, "filter_tags": {"sim_type": "driver"}},
            {
                "job_mode": "serial",
                "num_nodes": 127,
                "filter_tags": {"sim_type": "worker"},
            },
        ],
        description="Optionally, subdivide an allocation into multiple partitions.",
    )

    class Config:
        use_enum_values = True


class BatchJobCreate(BatchJobBase):
    site_id: int = Field(..., example=4, description="The Site id where this batchjob is submitted")
    project: str = Field(..., example="datascience", description="The project/allocation to charge for this batchjob")
    queue: str = Field(..., example="default", description="Which queue the batchjob is submitted on")


class BatchJobUpdate(BaseModel):
    scheduler_id: int = Field(None, example=14192, description="The local HPC scheduler's ID for this batchjob")
    state: BatchJobState = Field(
        None, example="queued", description="Status of this batchjob in the local HPC scheduler"
    )
    status_info: Dict[str, str] = Field(
        None, example={"error": "User is not a member of project X"}, description="Arbitrary status info"
    )
    start_time: Optional[datetime] = Field(None, description="BatchJob execution start time")
    end_time: Optional[datetime] = Field(None, description="BatchJob execution end time")


class BatchJobBulkUpdate(BatchJobUpdate):
    id: int = Field(..., example=8)


class BatchJobOut(BatchJobBase):
    id: int = Field(...)
    site_id: int = Field(...)
    scheduler_id: Optional[int] = Field(...)
    project: str = Field(...)
    queue: str = Field(...)
    state: BatchJobState = Field(...)
    status_info: Dict[str, str] = Field(...)
    start_time: Optional[datetime] = Field(...)
    end_time: Optional[datetime] = Field(...)

    class Config:
        orm_mode = True


class PaginatedBatchJobOut(BaseModel):
    count: int
    results: List[BatchJobOut]



from .app_run import AppRun, LocalAppRun
from .mpich import MPICHRun
from .openmpi import OpenMPIRun
from .perlmutter import PerlmutterGPURun
from .slurm import SlurmRun
from .summit import SummitJsrun
from .theta import ThetaAprun
from .theta_gpu import ThetaGPURun

__all__ = [
    "AppRun",
    "LocalAppRun",
    "ThetaAprun",
    "SlurmRun",
    "OpenMPIRun",
    "ThetaGPURun",
    "MPICHRun",
    "SummitJsrun",
    "PerlmutterGPURun",
]



from dataclasses import dataclass
from datetime import datetime
from typing import Dict, List, Set, Tuple, cast

from fastapi import Query
from sqlalchemy import orm

from balsam import schemas
from balsam.server.models import App, BatchJob, Job, LogEvent, Session, Site, TransferItem


@dataclass
class SiteQuery:
    name: str = Query(None, description="Fetch the site with this name")
    path: str = Query(None, description="Only return Sites with paths containing this string.")
    id: List[int] = Query(None, description="Only return Sites having an id in this list.")
    last_refresh_after: datetime = Query(None, description="Only return Sites active since this time (UTC)")

    def apply_filters(self, qs: "orm.Query[Site]") -> "orm.Query[Site]":
        if self.name:
            qs = qs.filter(Site.name == self.name)
        if self.path:
            qs = qs.filter(Site.path.like(f"%{self.path}%"))
        if self.id:
            qs = qs.filter(Site.id.in_(self.id))
        if self.last_refresh_after:
            qs = qs.filter(Site.last_refresh >= self.last_refresh_after)
        return qs


@dataclass
class AppQuery:
    site_id: List[int] = Query(None, description="Only return Apps associated with the Site IDs in this list.")
    id: List[int] = Query(None, description="Only return Apps with IDs in this list.")
    name: str = Query(None, description="Only return Apps having an ApplicationDefinition with this class name.")
    site_path: str = Query(None, description="Only return Apps from Sites having paths containing this substring.")
    site_name: str = Query(None, description="Only return Apps from the Site with this unique name.")

    def apply_filters(self, qs: "orm.Query[App]") -> "orm.Query[App]":
        if self.site_id:
            qs = qs.filter(App.site_id.in_(self.site_id))
        if self.id:
            qs = qs.filter(App.id.in_(self.id))
        if self.name is not None:
            qs = qs.filter(App.name == self.name)
        if self.site_path:
            qs = qs.filter(Site.path.like(f"%{self.site_path}%"))
        if self.site_name:
            qs = qs.filter(Site.name == self.site_name)
        return qs


@dataclass
class SessionQuery:
    id: List[int] = Query(None, description="Only return Sessions having an id in this list.")

    def apply_filters(self, qs: "orm.Query[Session]") -> "orm.Query[Session]":
        if self.id:
            qs = qs.filter(Session.id.in_(self.id))
        return qs


@dataclass
class EventLogQuery:
    id: List[int] = Query(None, description="Only return EventLogs having an id in this list.")
    job_id: List[int] = Query(None, description="Only return Events associated with Job IDs in this list.")
    batch_job_id: int = Query(None, description="Only return Events associated this BatchJob id.")
    scheduler_id: int = Query(None, description="Only return Events associated with this HPC scheduler job ID.")
    tags: List[str] = Query(
        None, description="Only return Events for Jobs containing these tags (list of KEY:VALUE strings)"
    )
    data: List[str] = Query(None, description="Only return Events containing this data (list of KEY:VALUE strings)")
    timestamp_before: datetime = Query(None, description="Only return Events before this time (UTC).")
    timestamp_after: datetime = Query(None, description="Only return Events that occured after this time (UTC).")
    from_state: str = Query(None, description="Only return Events transitioning from this Job state.")
    to_state: str = Query(None, description="Only return Events transitioning to this Job state.")
    ordering: schemas.EventOrdering = Query("-timestamp", description="Order events by this field.")

    def apply_filters(self, qs: "orm.Query[LogEvent]") -> "orm.Query[LogEvent]":
        if self.id:
            qs = qs.filter(LogEvent.id.in_(self.id))
        if self.job_id:
            qs = qs.filter(Job.id.in_(self.job_id))
        if self.batch_job_id or self.scheduler_id:
            qs = qs.join(BatchJob)  # type: ignore
        if self.batch_job_id:
            qs = qs.filter(Job.batch_job_id == self.batch_job_id)
        if self.scheduler_id:
            qs = qs.filter(BatchJob.scheduler_id == self.scheduler_id)
        if self.tags:
            tags_dict: Dict[str, str] = dict(t.split(":", 1) for t in self.tags if ":" in t)  # type: ignore
            qs = qs.filter(Job.tags.contains(tags_dict))  # type: ignore
        if self.data:
            data_dict: Dict[str, str] = dict(d.split(":", 1) for d in self.data if ":" in d)  # type: ignore
            qs = qs.filter(LogEvent.data.contains(data_dict))  # type: ignore
        if self.timestamp_before:
            qs = qs.filter(LogEvent.timestamp <= self.timestamp_before)
        if self.timestamp_after:
            qs = qs.filter(LogEvent.timestamp >= self.timestamp_after)
        if self.from_state:
            qs = qs.filter(LogEvent.from_state == self.from_state)
        if self.to_state:
            qs = qs.filter(LogEvent.to_state == self.to_state)
        if self.ordering:
            desc = self.ordering.startswith("-")
            order_col = getattr(LogEvent, self.ordering.lstrip("-"))
            qs = qs.order_by(order_col.desc() if desc else order_col)
        return qs


@dataclass
class JobQuery:
    id: List[int] = Query(None, min_items=1, description="Only return Jobs with ids in this list.")
    parent_id: List[int] = Query(
        None, min_items=1, description="Only return Jobs that are children of Jobs with ids in this list."
    )
    app_id: int = Query(None, description="Only return Jobs associated with this App id.")
    site_id: List[int] = Query(None, description="Only return Jobs associated with these Site ids.")
    batch_job_id: int = Query(None, description="Only return Jobs associated with this BatchJob id.")
    last_update_before: datetime = Query(
        None, description="Only return Jobs that were updated before this time (UTC)."
    )
    last_update_after: datetime = Query(None, description="Only return Jobs that were updated after this time (UTC).")
    workdir__contains: str = Query(None, description="Only return jobs with workdirs containing this string.")
    state__ne: schemas.JobState = Query(None, description="Only return jobs with states not equal to this state.")
    state: Set[schemas.JobState] = Query(None, description="Only return jobs in this set of states.")
    tags: List[str] = Query(None, description="Only return jobs containing these tags (list of KEY:VALUE strings)")
    pending_file_cleanup: bool = Query(None, description="Only return jobs which have not yet had workdir cleaned.")
    ordering: schemas.JobOrdering = Query(None, description="Order Jobs by this field.")

    def apply_filters(self, qs: "orm.Query[Job]") -> "orm.Query[Job]":  # noqa: C901
        if self.id:
            qs = qs.filter(Job.id.in_(self.id))
        if self.parent_id:
            qs = qs.filter(Job.parents.any(Job.id.in_(self.parent_id)))
        if self.app_id:
            qs = qs.filter(Job.app_id == self.app_id)
        if self.site_id:
            qs = qs.filter(Site.id.in_(self.site_id))
        if self.batch_job_id:
            qs: "orm.Query[Job]" = qs.join(BatchJob)  # type: ignore
            qs = qs.filter(BatchJob.id == self.batch_job_id)
        if self.last_update_before:
            qs = qs.filter(Job.last_update <= self.last_update_before)
        if self.last_update_after:
            qs = qs.filter(Job.last_update >= self.last_update_after)
        if self.workdir__contains:
            qs = qs.filter(Job.workdir.like(f"%{self.workdir__contains}%"))
        if self.state__ne:
            qs = qs.filter(Job.state != self.state__ne)
        if self.state:
            qs = qs.filter(Job.state.in_(self.state))
        if self.tags:
            tags_dict: Dict[str, str] = dict(t.split(":", 1) for t in self.tags if ":" in t)  # type: ignore
            qs = qs.filter(Job.tags.contains(tags_dict))  # type: ignore
        if self.pending_file_cleanup:
            qs = qs.filter(Job.pending_file_cleanup)
        if self.ordering:
            desc = self.ordering.startswith("-")
            order_col = getattr(Job, self.ordering.lstrip("-"))
            qs = qs.order_by(order_col.desc() if desc else order_col)
        return qs


@dataclass
class TransferItemQuery:
    id: List[int] = Query(None, description="Only return transfer items with IDs in this list.")
    site_id: int = Query(None, description="Only return transfer items associated with this Site id.")
    job_id: List[int] = Query(None, description="Only return transfer items associated with this Job id list.")
    state: Set[schemas.TransferItemState] = Query(
        None, description="Only return transfer items in this set of states."
    )
    direction: schemas.TransferDirection = Query(None, description="Only return items in this transfer direction.")
    job_state: str = Query(None, description="Only return transfer items for Jobs having this state.")
    tags: List[str] = Query(
        None, description="Only return transfer items for Jobs having these tags (list of KEY:VALUE strings)."
    )

    def apply_filters(self, qs: "orm.Query[TransferItem]") -> "orm.Query[TransferItem]":
        if self.id:
            qs = qs.filter(TransferItem.id.in_(self.id))
        if self.site_id:
            qs = qs.filter(Site.id == self.site_id)
        if self.job_id:
            qs = qs.filter(Job.id.in_(self.job_id))
        if self.state:
            qs = qs.filter(TransferItem.state.in_(self.state))
        if self.direction:
            qs = qs.filter(TransferItem.direction == self.direction)
        if self.job_state:
            qs = qs.filter(Job.state == self.job_state)
        if self.tags:
            tags_dict = dict(cast(Tuple[str, str], t.split(":", 1)) for t in self.tags if ":" in t)
            qs = qs.filter(Job.tags.contains(tags_dict))  # type: ignore
        return qs


@dataclass
class BatchJobQuery:
    id: List[int] = Query(None, description="Only return BatchJobs having an id in this list.")
    site_id: List[int] = Query(None, description="Only return batchjobs for Sites in this id list.")
    state: List[str] = Query(None, description="Only return batchjobs having one of these States in this list.")
    scheduler_id: int = Query(None, description="Return the batchjob with this local scheduler id.")
    queue: str = Query(None, description="Only return batchjobs submitted to this queue.")
    ordering: schemas.BatchJobOrdering = Query(None, description="Order batchjobs by this field.")
    start_time_before: datetime = Query(
        None, description="Only return batchjobs that started before this time (UTC)."
    )
    start_time_after: datetime = Query(None, description="Only return batchjobs that started after this time (UTC).")
    end_time_before: datetime = Query(None, description="Only return batchjobs that finished before this time (UTC).")
    end_time_after: datetime = Query(None, description="Only return batchjobs that finished after this time (UTC).")
    filter_tags: List[str] = Query(
        None, description="Only return batchjobs processing these tags (list of KEY:VALUE strings)."
    )

    def apply_filters(self, qs: "orm.Query[BatchJob]") -> "orm.Query[BatchJob]":
        if self.id:
            qs = qs.filter(BatchJob.id.in_(self.id))
        if self.site_id:
            qs = qs.filter(BatchJob.site_id.in_(self.site_id))
        if self.state:
            qs = qs.filter(BatchJob.state.in_(self.state))
        if self.scheduler_id:
            qs = qs.filter(BatchJob.scheduler_id == self.scheduler_id)
        if self.queue:
            qs = qs.filter(BatchJob.queue == self.queue)

        if self.start_time_before:
            qs = qs.filter(BatchJob.start_time <= self.start_time_before)
        if self.start_time_after:
            qs = qs.filter(BatchJob.start_time >= self.start_time_after)

        if self.end_time_before:
            qs = qs.filter(BatchJob.end_time <= self.end_time_before)
        if self.end_time_after:
            qs = qs.filter(BatchJob.end_time >= self.end_time_after)

        if self.filter_tags:
            tags_dict: Dict[str, str] = dict(t.split(":", 1) for t in self.filter_tags if ":" in t)  # type: ignore
            qs = qs.filter(BatchJob.filter_tags.contains(tags_dict))  # type: ignore
        if self.ordering:
            desc = self.ordering.startswith("-")
            order_col = getattr(BatchJob, self.ordering.lstrip("-"))
            qs = qs.order_by(order_col.desc() if desc else order_col)
        return qs



from uuid import uuid4

from fastapi import status

from balsam.client import urls


def test_unauth_user_cannot_view_sites(anon_client):
    anon_client.get("/sites/", check=status.HTTP_401_UNAUTHORIZED)


def test_register(anon_client):
    login_credentials = {"username": f"user{uuid4()}", "password": "foo"}
    resp = anon_client.post("/" + urls.PASSWORD_REGISTER, **login_credentials)
    assert type(resp["id"]) == int
    assert resp["username"] == login_credentials["username"]


def test_auth_user_can_view_sites(auth_client):
    resp = auth_client.get("/sites/")
    assert resp["results"] == []



import random
from datetime import datetime, timedelta

from fastapi import status

from .util import create_site


def test_can_create_batchjob(auth_client):
    site1 = create_site(auth_client, name="site1")
    batch_job = auth_client.post(
        "/batch-jobs/",
        site_id=site1["id"],
        project="datascience",
        queue="default",
        num_nodes=4,
        wall_time_min=60,
        job_mode="mpi",
        optional_params={},
        filter_tags={},
    )
    assert batch_job["state"] == "pending_submission"
    assert batch_job["scheduler_id"] is None


def test_list_batchjobs_spanning_sites(auth_client):
    site1 = create_site(auth_client, name="site1")
    site2 = create_site(auth_client, name="site2")
    for time in [10, 20, 30, 40]:
        for site in [site1, site2]:
            auth_client.post(
                "/batch-jobs/",
                site_id=site["id"],
                project="datascience",
                queue="default",
                num_nodes=4,
                wall_time_min=time,
                job_mode="mpi",
                optional_params={},
                filter_tags={},
            )
    bjob_list = auth_client.get("/batch-jobs")
    assert bjob_list["count"] == 8


def test_filter_by_site(auth_client):
    site1 = create_site(auth_client, name="site1")
    site2 = create_site(auth_client, name="site2")
    for time in [10, 20, 30, 40]:
        for site in [site1, site2]:
            auth_client.post(
                "/batch-jobs/",
                site_id=site["id"],
                project="datascience",
                queue="default",
                num_nodes=4,
                wall_time_min=time,
                job_mode="mpi",
                optional_params={},
                filter_tags={},
            )
    # providing GET kwargs causes result list to be filtered
    bjob_list = auth_client.get("/batch-jobs", site_id=site2["id"])
    assert bjob_list["count"] == 4
    for bjob in bjob_list["results"]:
        assert bjob["site_id"] == site2["id"]


def test_filter_by_time_range(auth_client):
    site = create_site(auth_client, name="site1")
    # Create 10 historical batchjobs
    # Job n started n hours ago and took 30 minutes
    now = datetime.utcnow()
    for i in range(1, 11):
        start = now - timedelta(hours=i * 1)
        end = start + timedelta(minutes=30)
        if now - timedelta(hours=5) <= end <= now - timedelta(hours=3):
            filter_tags = {"good": "Yes"}
        else:
            filter_tags = {}
        bjob = auth_client.post(
            "/batch-jobs/",
            site_id=site["id"],
            project="datascience",
            queue="default",
            num_nodes=4,
            wall_time_min=30,
            job_mode="mpi",
            filter_tags=filter_tags,
        )
        auth_client.put(
            f"/batch-jobs/{bjob['id']}",
            state="finished",
            start_time=start,
            end_time=end,
        )

    # Now, we want to filter for jobs that ended between 3 and 5 hours ago
    # The end_times are: 0.5h ago, 1.5 ago, 2.5, 3.5, 4.5, 5.5, ...
    # So we should have 2 jobs land in this window
    end_after = now - timedelta(hours=5)
    end_before = now - timedelta(hours=3)
    jobs = auth_client.get("/batch-jobs/", end_time_after=end_after, end_time_before=end_before)
    assert jobs["count"] == 2 == len(jobs["results"])
    for job in jobs["results"]:
        assert "good" in job["filter_tags"]


def test_json_tags_filter_list(auth_client):
    site = create_site(auth_client, name="site1")
    for priority in [None, 1, 2, 3]:
        for system in ["H2O", "D2O", "HF"]:
            if priority:
                tags = {"priority": priority, "system": system}
            else:
                tags = {"system": system}
            auth_client.post(
                "/batch-jobs/",
                site_id=site["id"],
                project="datascience",
                queue="default",
                num_nodes=4,
                wall_time_min=30,
                job_mode="mpi",
                filter_tags=tags,
            )

    jobs = auth_client.get("/batch-jobs/")
    assert jobs["count"] == 12 == len(jobs["results"])

    jobs = auth_client.get("/batch-jobs/", filter_tags="priority:2")
    assert jobs["count"] == 3 == len(jobs["results"])

    jobs = auth_client.get("/batch-jobs/", filter_tags="system:D2O")
    assert jobs["count"] == 4 == len(jobs["results"])

    jobs = auth_client.get("/batch-jobs/", filter_tags=["system:D2O", "priority:1"])
    assert jobs["count"] == 1 == len(jobs["results"])


def test_detail_view(auth_client):
    site = create_site(auth_client, name="site1")
    bjob = auth_client.post(
        "/batch-jobs/",
        site_id=site["id"],
        project="datascience",
        queue="default",
        num_nodes=4,
        wall_time_min=30,
        job_mode="mpi",
    )
    id = bjob["id"]
    retrieved = auth_client.get(f"/batch-jobs/{id}")
    assert retrieved == bjob


def test_update_to_invalid_state(auth_client):
    site = create_site(auth_client, name="site1")
    bjob = auth_client.post(
        "/batch-jobs/",
        site_id=site["id"],
        project="datascience",
        queue="default",
        num_nodes=4,
        wall_time_min=30,
        job_mode="mpi",
    )
    id = bjob["id"]
    auth_client.put(
        f"batch-jobs/{id}",
        state="invalid-state",
        check=status.HTTP_422_UNPROCESSABLE_ENTITY,
    )


def test_update_valid(auth_client):
    site = create_site(auth_client, name="site1")
    bjob = auth_client.post(
        "/batch-jobs/",
        site_id=site["id"],
        project="datascience",
        queue="default",
        num_nodes=4,
        wall_time_min=30,
        job_mode="mpi",
    )
    id = bjob["id"]
    auth_client.put(
        f"batch-jobs/{id}",
        state="submit_failed",
        status_info={"error": "User is not a member of project datascience"},
    )
    ret = auth_client.get(f"/batch-jobs/{id}")
    assert ret["state"] == "submit_failed"
    assert "error" in ret["status_info"]


def test_bulk_status_update_batch_jobs(auth_client):
    theta = create_site(auth_client, name="theta")
    cooley = create_site(auth_client, name="cooley")
    for _ in range(10):
        auth_client.post(
            "/batch-jobs/",
            site_id=theta["id"],
            project="datascience",
            queue="default",
            num_nodes=4,
            wall_time_min=30,
            job_mode="mpi",
        )
        auth_client.post(
            "/batch-jobs/",
            site_id=cooley["id"],
            project="datascience",
            queue="default",
            num_nodes=4,
            wall_time_min=30,
            job_mode="mpi",
        )

    # scheduler agent receives 10 batchjobs; sends back bulk-state updates
    jobs = auth_client.get("/batch-jobs/", site_id=cooley["id"])
    assert jobs["count"] == 10
    jobs = jobs["results"]
    for job in jobs[:5]:
        job["state"] = "queued"
    for job in jobs[5:]:
        job["state"] = "running"
        job["start_time"] = datetime.utcnow() + timedelta(minutes=random.randint(-30, 0))

    updates = [{k: job[k] for k in job if k in ["id", "state", "start_time"]} for job in jobs]
    result = auth_client.bulk_patch("/batch-jobs/", updates)

    for updated_job in result:
        id = updated_job["id"]
        expected_state = next(j["state"] for j in jobs if j["id"] == id)
        assert updated_job["state"] == expected_state

    jobs = auth_client.get("/batch-jobs/", site_id=cooley["id"], state="running")
    assert jobs["count"] == len(jobs["results"]) == 5


def test_delete_running_batchjob(auth_client):
    site = create_site(auth_client, name="theta")
    bjob = auth_client.post(
        "/batch-jobs/",
        site_id=site["id"],
        project="datascience",
        queue="default",
        num_nodes=7,
        wall_time_min=30,
        job_mode="mpi",
    )
    id = bjob["id"]
    user_job = auth_client.get(f"/batch-jobs/{id}")
    site_job = auth_client.get(f"/batch-jobs/{id}")

    # site updates to running
    site_job = auth_client.put(
        f"/batch-jobs/{id}",
        scheduler_id=123,
        start_time=datetime.utcnow(),
        state="running",
    )
    assert site_job["state"] == "running"
    assert site_job["scheduler_id"] == 123

    # user patches to pending_deletion
    user_job = auth_client.put(f"/batch-jobs/{id}", state="pending_deletion")
    assert user_job["state"] == "pending_deletion"

    # Client receives job marked for deletion
    site_job = auth_client.get(f"/batch-jobs/{id}")
    assert site_job["state"] == "pending_deletion"
    patch = {
        "id": id,
        "state": "finished",
        "end_time": datetime.utcnow(),
        "status_info": {"message": "User deleted job"},
    }
    auth_client.bulk_patch("/batch-jobs/", [patch])

    # User sees the deletion eventually
    user_job = auth_client.get(f"/batch-jobs/{id}")
    assert user_job["state"] == "finished"
    assert "User deleted job" in user_job["status_info"].values()


def test_delete_api_endpoint(auth_client):
    site = create_site(auth_client, name="theta")
    bjob = auth_client.post(
        "/batch-jobs/",
        site_id=site["id"],
        project="datascience",
        queue="default",
        num_nodes=7,
        wall_time_min=30,
        job_mode="mpi",
    )
    all = auth_client.get("/batch-jobs/")
    assert all["count"] == 1
    resp = auth_client.delete(f"/batch-jobs/{bjob['id']}")
    print("delete response is", resp)
    all = auth_client.get("/batch-jobs/")
    assert all["count"] == 0


def test_no_shared_batchjobs_in_list_view(fastapi_user_test_client):
    """client2 cannot see client1's batchjobs"""
    client1, client2 = fastapi_user_test_client(), fastapi_user_test_client()
    site1 = create_site(client1, path="/bar")
    create_site(client2, path="/foo")

    # client1 adds batchjob to site1
    client1.post(
        "/batch-jobs/",
        site_id=site1["id"],
        project="datascience",
        queue="default",
        num_nodes=7,
        wall_time_min=30,
        job_mode="mpi",
    )
    # client2 cannot see it
    assert client2.get("/batch-jobs")["count"] == 0
    assert client1.get("/batch-jobs")["count"] == 1


def test_permission_in_detail_view(fastapi_user_test_client):
    """client2 cannot see client1's batchjobs in detail view"""
    client1, client2 = fastapi_user_test_client(), fastapi_user_test_client()
    site1 = create_site(client1, path="/bar")
    create_site(client2, path="/foo")
    bjob = client1.post(
        "/batch-jobs/",
        site_id=site1["id"],
        project="datascience",
        queue="default",
        num_nodes=7,
        wall_time_min=30,
        job_mode="mpi",
    )
    id = bjob["id"]
    # client2 gets 404 if try to access client1's bjob
    client2.get(f"/batch-jobs/{id}", check=status.HTTP_404_NOT_FOUND)
    client1.get(f"/batch-jobs/{id}", check=status.HTTP_200_OK)


def test_bulk_update_cannot_affect_other_users_batchjobs(fastapi_user_test_client):
    """client2 bulk-update cannot affect client1's batchjobs"""
    client1, client2 = fastapi_user_test_client(), fastapi_user_test_client()
    site1 = create_site(client1, path="/bar")
    create_site(client2, path="/foo")
    bjob = client1.post(
        "/batch-jobs/",
        site_id=site1["id"],
        project="datascience",
        queue="default",
        num_nodes=7,
        wall_time_min=30,
        job_mode="mpi",
    )

    # client 2 attempts bulk update with client1's batchjob id; fails
    patch = {"id": bjob["id"], "state": "pending_deletion"}
    client2.bulk_patch("/batch-jobs/", [patch], check=status.HTTP_400_BAD_REQUEST)

    # client 1 can do it, though:
    client1.bulk_patch("/batch-jobs/", [patch], check=status.HTTP_200_OK)



from .apps import AppCreate, AppOut, AppParameter, AppUpdate, PaginatedAppsOut, TransferSlot
from .batchjob import (
    BatchJobBulkUpdate,
    BatchJobCreate,
    BatchJobOrdering,
    BatchJobOut,
    BatchJobPartition,
    BatchJobState,
    BatchJobUpdate,
    JobMode,
    PaginatedBatchJobOut,
    SchedulerBackfillWindow,
    SchedulerJobLog,
    SchedulerJobStatus,
)
from .job import (
    DONE_STATES,
    RUNNABLE_STATES,
    JobBulkUpdate,
    JobCreate,
    JobOrdering,
    JobOut,
    JobState,
    JobTransferItem,
    JobUpdate,
    PaginatedJobsOut,
    ServerJobCreate,
)
from .logevent import EventOrdering, LogEventOut, PaginatedLogEventOut
from .serializer import (
    DeserializeError,
    EmptyPayload,
    SerializeError,
    deserialize,
    get_source,
    raise_from_serialized,
    serialize,
    serialize_exception,
)
from .session import PaginatedSessionsOut, SessionAcquire, SessionCreate, SessionOut
from .site import AllowedQueue, PaginatedSitesOut, SiteCreate, SiteOut, SiteUpdate
from .transfer import (
    PaginatedTransferItemOut,
    TransferDirection,
    TransferItemBulkUpdate,
    TransferItemOut,
    TransferItemState,
    TransferItemUpdate,
)
from .user import UserCreate, UserOut

__all__ = [
    "UserCreate",
    "UserOut",
    "SiteCreate",
    "SiteUpdate",
    "SiteOut",
    "PaginatedSitesOut",
    "AllowedQueue",
    "AppCreate",
    "AppUpdate",
    "AppOut",
    "PaginatedAppsOut",
    "AppParameter",
    "TransferSlot",
    "BatchJobCreate",
    "BatchJobUpdate",
    "BatchJobBulkUpdate",
    "BatchJobState",
    "BatchJobPartition",
    "BatchJobOut",
    "BatchJobOrdering",
    "JobMode",
    "PaginatedBatchJobOut",
    "SessionCreate",
    "SessionOut",
    "PaginatedSessionsOut",
    "SessionAcquire",
    "JobCreate",
    "ServerJobCreate",
    "JobUpdate",
    "JobBulkUpdate",
    "PaginatedJobsOut",
    "JobOut",
    "JobState",
    "JobOrdering",
    "JobTransferItem",
    "RUNNABLE_STATES",
    "DONE_STATES",
    "TransferItemOut",
    "PaginatedTransferItemOut",
    "TransferItemUpdate",
    "TransferItemBulkUpdate",
    "TransferItemState",
    "TransferDirection",
    "LogEventOut",
    "PaginatedLogEventOut",
    "EventOrdering",
    "SchedulerBackfillWindow",
    "SchedulerJobLog",
    "SchedulerJobStatus",
    "serialize",
    "deserialize",
    "serialize_exception",
    "raise_from_serialized",
    "get_source",
    "SerializeError",
    "DeserializeError",
    "EmptyPayload",
]



import logging
import random
import time
from datetime import datetime, timedelta
from pathlib import Path
from typing import Any, Callable, Dict, List, TextIO

import click
import yaml
from pydantic import BaseSettings, validator

from balsam.api import App, Job, JobState, Site
from balsam.schemas import RUNNABLE_STATES

logfmt = "%(asctime)s.%(msecs)03d | %(levelname)s | %(lineno)s] %(message)s"
logging.basicConfig(filename=None, level=logging.DEBUG, format=logfmt, datefmt="%Y-%m-%d %H:%M:%S", force=True)
logger = logging.getLogger()


class XPCS(BaseSettings):
    result_dir: Path
    remote_alias: str
    h5_in: Path
    imm_in: Path


class Eig(BaseSettings):
    remote_alias: str
    result_dir: Path
    matrix_in: Path


class ExperimentConfig(BaseSettings):
    experiment_tag: str
    submission_mode: str
    submit_period: float
    submit_batch_size: int
    max_site_backlog: int
    experiment_duration_min: int
    site_ids: List[int]
    app_name: str
    site_cpu_map: Dict[int, int]

    xpcs_datasets: List[XPCS]
    eig_datasets: List[Eig]

    @validator("submission_mode")
    def valid_submit_mode(cls, v: str) -> str:
        if v not in ["const-backlog", "shortest-backlog", "round-robin"]:
            raise ValueError(f"invalid mode: {v}")
        return v


class JobFactory:
    def __init__(
        self,
        experiment_tag: str,
        xpcs_datasets: List[XPCS],
        eig_datasets: List[Eig],
        site_cpu_map: Dict[int, int],
    ) -> None:
        self.idx = 0
        self.submission_idx = 0
        self.experiment_tag = experiment_tag
        self.xpcs_datasets = xpcs_datasets
        self.eig_datasets = eig_datasets
        self.generators: Dict[str, Callable[..., Job]] = {
            "xpcs.EigenCorr": self.xpcs_eigen,
            "eig.Eig": self.eig,
        }
        self.site_cpu_map = site_cpu_map

    def submit_jobs(self, app: App, num_jobs: int) -> List[Job]:
        job_factory = self.generators[app.class_path]
        jobs = [job_factory(app) for _ in range(num_jobs)]
        Job.objects.bulk_create(jobs)
        self.submission_idx += 1
        return jobs

    def xpcs_eigen(self, app: App) -> Job:
        assert app.id is not None

        transfer_set = random.choice(self.xpcs_datasets)
        source_tag = transfer_set.remote_alias

        workdir = Path(f"{self.experiment_tag}/{source_tag}/corr_{self.idx:06d}")
        result_path = transfer_set.result_dir.joinpath(transfer_set.h5_in.name).with_suffix(
            f".result{self.idx:06d}.hdf"
        )
        num_cpus = self.site_cpu_map[app.site_id]

        transfers: Dict[str, Any] = {
            "h5_in": {
                "location_alias": transfer_set.remote_alias,
                "path": transfer_set.h5_in,
            },
            "imm_in": {
                "location_alias": transfer_set.remote_alias,
                "path": transfer_set.imm_in,
            },
            "h5_out": {
                "location_alias": transfer_set.remote_alias,
                "path": result_path,
            },
        }
        job = Job(
            workdir=workdir,
            app_id=app.id,
            num_nodes=1,
            node_packing_count=1,
            threads_per_rank=num_cpus,
            transfers=transfers,
            tags={"job_source": source_tag, "experiment": self.experiment_tag},
        )
        self.idx += 1
        return job

    def eig(self, app: App) -> Job:
        assert app.id is not None

        transfer_set = random.choice(self.eig_datasets)
        source_tag = transfer_set.remote_alias

        workdir = Path(f"{self.experiment_tag}/{source_tag}/eig_{self.idx:06d}")
        result_path = transfer_set.result_dir.joinpath(transfer_set.matrix_in.name).with_suffix(
            f".eig{self.idx:06d}.npy"
        )

        transfers: Dict[str, Any] = {
            "matrix": {
                "location_alias": transfer_set.remote_alias,
                "path": transfer_set.matrix_in,
            },
            "eigvals": {
                "location_alias": transfer_set.remote_alias,
                "path": result_path,
            },
        }
        job = Job(
            workdir=workdir,
            app_id=app.id,
            num_nodes=1,
            node_packing_count=1,
            threads_per_rank=16,
            transfers=transfers,
            tags={"job_source": source_tag, "experiment": self.experiment_tag},
        )
        self.idx += 1
        return job


def get_site_backlogs(site_ids: List[int]) -> Dict[int, int]:
    backlogs: Dict[int, int] = {}
    for site_id in site_ids:
        qs = Job.objects.filter(site_id=site_id, state=set([*RUNNABLE_STATES, JobState.ready, JobState.staged_in]))
        count = qs.count()
        assert count is not None
        backlogs[site_id] = count
    return backlogs


def submit_const_backlog(
    job_factory: JobFactory,
    apps_by_site: Dict[int, App],
    backlogs_by_site: Dict[int, int],
    batch_size: int,
    max_backlog: int,
) -> None:
    for site_id, app in apps_by_site.items():
        backlog = backlogs_by_site[site_id]
        num_submit = min(batch_size, max_backlog - backlog)
        if num_submit < 1:
            logger.info(f"Site {site_id} at max_backlog: skipping")
        else:
            job_factory.submit_jobs(app, num_submit)
            logger.info(f"Submitted {num_submit} to Site {site_id}")


def submit_shortest_backlog(
    job_factory: JobFactory,
    apps_by_site: Dict[int, App],
    backlogs_by_site: Dict[int, int],
    batch_size: int,
    max_backlog: int,
) -> None:
    # Select Site with smallest backlog
    submit_site_id, backlog = min(backlogs_by_site.items(), key=lambda x: x[1])
    app = apps_by_site[submit_site_id]
    if backlog < max_backlog:
        job_factory.submit_jobs(app, batch_size)
        logger.info(f"Submitted {batch_size} jobs to Site {submit_site_id}")
    else:
        logger.info(f"Will not submit new jobs; at max backlog: {backlog} / {max_backlog}")


def submit_round_robin(
    job_factory: JobFactory,
    apps_by_site: Dict[int, App],
    backlogs_by_site: Dict[int, int],
    batch_size: int,
    max_backlog: int,
) -> None:
    # Select next Site in turn
    site_idx = job_factory.submission_idx % len(apps_by_site)
    site_id = list(apps_by_site.keys())[site_idx]
    app = apps_by_site[site_id]
    backlog = backlogs_by_site[site_id]
    logger.info(f"Backlogs by site: {backlogs_by_site}")
    if backlog < max_backlog:
        job_factory.submit_jobs(app, batch_size)
        logger.info(f"Submitted {batch_size} jobs to Site {site_id}")
    else:
        job_factory.submission_idx += 1
        logger.info(f"Will not submit new jobs to Site {site_id}; at max backlog: {backlog} / {max_backlog}")


@click.command()
@click.option("-c", "--config-file", required=True, type=click.File("r"))
def main(config_file: TextIO) -> None:
    config = ExperimentConfig(**yaml.safe_load(config_file))
    logger.debug(f"Loaded experiment config: {yaml.dump(config.dict(), sort_keys=False, indent=2)}")

    site_names = {site.id: (site.name, site.path.name) for site in Site.objects.filter(id=config.site_ids)}
    assert len(site_names) == len(
        config.site_ids
    ), f"Config specified site_ids {config.site_ids} but API only found {len(site_names)} of them."

    apps_by_site = {
        site_id: App.objects.get(site_id=site_id, class_path=config.app_name) for site_id in config.site_ids
    }

    job_factory = JobFactory(
        config.experiment_tag,
        config.xpcs_datasets,
        config.eig_datasets,
        config.site_cpu_map,
    )

    if config.submission_mode == "const-backlog":
        submit_method = submit_const_backlog
    elif config.submission_mode == "round-robin":
        submit_method = submit_round_robin
    elif config.submission_mode == "shortest-backlog":
        submit_method = submit_shortest_backlog
    else:
        raise ValueError("Invalid submission mode")

    start = datetime.utcnow()
    logger.info(f"Starting experiment at {start}")
    logger.info(f"Total duration will be {config.experiment_duration_min} minutes at most")

    while datetime.utcnow() - start < timedelta(minutes=config.experiment_duration_min):
        time.sleep(config.submit_period)
        backlogs = get_site_backlogs(config.site_ids)
        submit_method(
            job_factory,
            apps_by_site,
            backlogs,
            config.submit_batch_size,
            config.max_site_backlog,
        )

    logger.info("Reached experiment max duration, exiting.")


if __name__ == "__main__":
    main()



import sys
import tempfile

from balsam.config import Settings, site_builder

if __name__ == "__main__":
    for path, site_defaults in site_builder.load_default_configs().items():
        try:
            with tempfile.NamedTemporaryFile(mode="w") as fp:
                txt = site_builder.render_settings_file(site_defaults.dict())
                fp.write(txt)
                fp.flush()
                Settings.load(fp.name)
        except Exception as exc:
            print(f"Invalid default settings in {path}\n  --> {exc}")
            sys.exit(1)
        else:
            print(f"{path.name}: OK")



from balsam.api import ApplicationDefinition


class EigenCorr(ApplicationDefinition):
    """
    Runs XPCS-Eigen on an (H5, IMM) file pair from a remote location.

    Jobs do not require any command template parameters; instead,
    users provide `h5_in`, `imm_in`, and `h5_out` transfer items.
    """

    site = 0
    corr_exe = "/projects/datascience/aps/software/xpcs-eigen2/build/corr"
    command_template = f"{corr_exe} inp.h5 -imm inp.imm"
    environment_variables = {
        "HDF5_USE_FILE_LOCKING": "FALSE",
    }
    parameters = {}
    cleanup_files = ["*.hdf", "*.imm", "*.h5"]
    transfers = {
        "h5_in": {
            "required": True,
            "direction": "in",
            "local_path": "inp.h5",
            "description": "Input HDF5 file",
            "recursive": False,
        },
        "imm_in": {
            "required": True,
            "direction": "in",
            "local_path": "inp.imm",
            "description": "Input IMM file",
            "recursive": False,
        },
        "h5_out": {
            "required": True,
            "direction": "out",
            "local_path": "inp.h5",  # output is input, modified in-place
            "description": "Output H5 file",
            "recursive": False,
        },
    }



from datetime import datetime
from typing import Tuple, Union

import jwt
from fastapi import Depends, HTTPException, status
from fastapi.security import OAuth2PasswordBearer
from jwt import PyJWTError

from balsam import schemas
from balsam.server import settings

oauth2_scheme = OAuth2PasswordBearer(tokenUrl="auth/login/password")


def create_access_token(user: schemas.UserOut) -> Tuple[Union[bytes, str], datetime]:
    expiry = datetime.utcnow() + settings.auth.token_ttl
    to_encode = {"sub": user.id, "exp": expiry, "username": user.username}
    encoded_jwt = jwt.encode(to_encode, settings.auth.secret_key, algorithm=settings.auth.algorithm)
    return encoded_jwt, expiry


def user_from_token(token: str = Depends(oauth2_scheme)) -> schemas.UserOut:
    credentials_exception = HTTPException(
        status_code=status.HTTP_401_UNAUTHORIZED,
        detail="Could not validate credentials",
        headers={"WWW-Authenticate": "Bearer"},
    )
    expired_token_exc = HTTPException(
        status_code=status.HTTP_401_UNAUTHORIZED,
        detail="Expired Token",
        headers={"WWW-Authenticate": "Bearer"},
    )
    try:
        payload = jwt.decode(
            token,
            settings.auth.secret_key,
            algorithms=[settings.auth.algorithm],
        )

        user_id: int = payload["sub"]
        username: str = payload["username"]
        expiry = payload["exp"]

        utc_now = (datetime.utcnow() - datetime.utcfromtimestamp(0)).total_seconds()
        if expiry <= utc_now:
            raise expired_token_exc

        if user_id is None or username is None:
            raise credentials_exception

    except PyJWTError:
        raise credentials_exception

    return schemas.UserOut(id=user_id, username=username)



import logging
import os
import subprocess
import tempfile
from abc import ABC, abstractmethod
from pathlib import Path
from typing import IO, Dict, List, Optional, Union, cast

import psutil  # type: ignore

from balsam.site.launcher import NodeSpec

logger = logging.getLogger(__name__)


class _MockProcess:
    def cpu_affinity(self, cpus: Optional[List[int]] = None) -> None:
        pass


_psutil_process = psutil.Process()
try:
    _psutil_process.cpu_affinity()
except AttributeError:
    _psutil_process = _MockProcess()
    logger.debug("No psutil cpu_affinity support")


class TimeoutExpired(subprocess.TimeoutExpired):
    pass


class AppRun(ABC):
    """
    Balsam-wide Application launch interface
    The API defines the spec of an App invocation and manages
    its lifecycle: start/poll/kill/tail_output, etc...
    """

    def __init__(
        self,
        cmdline: str,
        preamble: Union[None, str, List[str]],
        envs: Dict[str, str],
        cwd: Path,
        outfile_path: Path,
        node_spec: NodeSpec,
        ranks_per_node: int,
        threads_per_rank: int,
        threads_per_core: int,
        launch_params: Dict[str, str],
        gpus_per_rank: int,
    ):
        self._cmdline = cmdline
        self._preamble = preamble
        self._envs = envs
        self._cwd = cwd
        self._outfile_path = outfile_path
        self._node_spec = node_spec
        self._ranks_per_node = ranks_per_node
        self._threads_per_rank = threads_per_rank
        self._threads_per_core = threads_per_core
        self._launch_params = launch_params
        self._gpus_per_rank = gpus_per_rank

    def get_num_ranks(self) -> int:
        return self._ranks_per_node * len(self._node_spec.node_ids)

    def get_cpus_per_rank(self) -> int:
        cpu_per_rank = len(self._node_spec.cpu_ids[0]) // self._ranks_per_node
        if not cpu_per_rank:
            cpu_per_rank = max(1, int(self._threads_per_rank // self._threads_per_core))
        return cpu_per_rank

    @abstractmethod
    def start(self) -> None:
        pass

    @abstractmethod
    def poll(self) -> Optional[int]:
        pass

    @abstractmethod
    def terminate(self) -> None:
        pass

    @abstractmethod
    def kill(self) -> None:
        pass

    @abstractmethod
    def tail_output(self, nlines: int = 10) -> str:
        pass

    @abstractmethod
    def wait(self, timeout: Optional[float] = None) -> int:
        pass


class FailedStartProcess:
    def poll(self) -> Optional[int]:
        return 12345

    def terminate(self) -> None:
        pass

    def kill(self) -> None:
        pass

    def wait(self, timeout: Optional[float] = None) -> int:
        return 12345


class SubprocessAppRun(AppRun):
    """
    Implements subprocess management for apps launched via Popen
    """

    _preamble_cache: Dict[str, Path] = {}

    def _build_cmdline(self) -> str:
        return ""

    def _build_preamble(self) -> str:
        if not self._preamble:
            return ""
        if isinstance(self._preamble, list):
            return " && ".join(self._preamble) + " && "

        if self._preamble not in self._preamble_cache:
            with tempfile.NamedTemporaryFile(mode="w", delete=False) as fp:
                fp.write(self._preamble)
                fp.flush()
                self._preamble_cache[self._preamble] = Path(fp.name).resolve()
        return f"source {self._preamble_cache[self._preamble]} && "

    def _set_envs(self) -> None:
        envs = os.environ.copy()
        envs.update(self._envs)
        # Check the assigned GPU ID list from the first compute node:
        gpu_ids = self._node_spec.gpu_ids[0]
        if gpu_ids:
            envs["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
            envs["CUDA_VISIBLE_DEVICES"] = ",".join(map(str, gpu_ids))
        envs["OMP_NUM_THREADS"] = str(self._threads_per_rank)
        self._envs = envs

    def _open_outfile(self) -> IO[bytes]:
        return open(self._outfile_path, "wb")

    def _pre_popen(self) -> None:
        pass

    def _post_popen(self) -> None:
        pass

    def start(self) -> None:
        self._set_envs()
        cmdline = self._build_preamble() + self._build_cmdline()
        logger.info(f"{self.__class__.__name__} Popen: {cmdline}")
        self._outfile = self._open_outfile()
        self._pre_popen()

        try:
            self._process = subprocess.Popen(
                cmdline,
                shell=True,
                executable="/bin/bash",
                stdout=self._outfile,
                stderr=subprocess.STDOUT,
                stdin=subprocess.DEVNULL,
                env=self._envs,
                cwd=self._cwd,
            )
        except Exception as e:
            logger.error(f"Popen failed: {e}")
            self._process = cast("subprocess.Popen[bytes]", FailedStartProcess())
        self._post_popen()

    def poll(self) -> Optional[int]:
        returncode = self._process.poll()
        if returncode is not None:
            self._outfile.close()
        return returncode

    def terminate(self) -> None:
        return self._process.terminate()

    def kill(self) -> None:
        self._process.kill()
        self._outfile.close()
        self._process.poll()

    def wait(self, timeout: Optional[float] = None) -> int:
        try:
            retcode = self._process.wait(timeout=timeout)
        except subprocess.TimeoutExpired:
            cmdline = self._build_preamble() + self._build_cmdline()
            assert timeout is not None
            raise TimeoutExpired(cmd=cmdline, timeout=timeout)
        else:
            self._outfile.close()
            return retcode

    def tail_output(self, nlines: int = 10) -> str:
        return subprocess.check_output(
            ["tail", "-n", str(nlines), self._outfile_path],
            encoding="utf-8",
        )


class LocalAppRun(SubprocessAppRun):
    def _build_cmdline(self) -> str:
        return self._cmdline

    def _pre_popen(self) -> None:
        cpu_ids = self._node_spec.cpu_ids[0]
        if cpu_ids:
            _psutil_process.cpu_affinity(cpu_ids)

    def _post_popen(self) -> None:
        _psutil_process.cpu_affinity([])



import itertools
import logging
import logging.handlers
import os
import socket
import sys
import threading
import time
from pathlib import Path
from typing import Any, Optional, TextIO, Union

import multiprocessing_logging  # type: ignore

from .sighandler import SigHandler


class PeriodicMemoryHandler(logging.handlers.MemoryHandler):
    """
    Buffer logs in memory; flush when:
        - urgency exceeds flushLevel (see errors right away)
        - logs exceed buffer capacity
        - flush_period seconds have elapsed
    A background thread is responsible for periodically flushing when the
    buffer is sitting idle.
    """

    def __init__(
        self,
        capacity: int,
        target: logging.Handler,
        flushLevel: int = logging.ERROR,
        flushOnClose: bool = True,
        flush_period: int = 30,
    ) -> None:
        super().__init__(
            capacity,
            flushLevel=flushLevel,
            target=target,
            flushOnClose=flushOnClose,
        )
        self.flush_period = flush_period
        self.last_flush = 0.0
        self.flushLevel = flushLevel
        self.target = target
        self.capacity = capacity
        self.flushOnClose = flushOnClose

    def flush(self) -> None:
        super().flush()
        self.last_flush = time.time()

    def shouldFlush(self, record: logging.LogRecord) -> bool:
        """
        Check for buffer full or a record at the flushLevel or higher.
        """
        return (
            (len(self.buffer) >= self.capacity)
            or (record.levelno >= self.flushLevel)
            or (time.time() - self.last_flush > self.flush_period)
        )


def validate_log_level(level: Union[str, int]) -> int:
    try:
        level = int(level)
    except ValueError:
        level = getattr(logging, str(level), logging.DEBUG)
        return int(level)
    return min(50, max(0, level))


def config_root_logger(level: Union[str, int, None] = None) -> logging.Logger:
    if level is None:
        level = validate_log_level(os.environ.get("BALSAM_LOG_LEVEL", "WARNING"))
    logger = logging.getLogger("balsam")
    logger.setLevel(level)
    if logger.hasHandlers():
        logger.handlers.clear()

    handler = logging.StreamHandler()
    formatter = logging.Formatter("%(levelname)s|%(name)s:%(lineno)s] %(message)s")
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger


def config_file_logging(
    filename: Union[str, Path],
    level: int,
    format: str,
    datefmt: str,
    buffer_num_records: int,
    flush_period: int,
) -> None:
    level = validate_log_level(level)
    root_logger = config_root_logger(level)

    file_handler = logging.FileHandler(filename=filename)
    mem_handler = PeriodicMemoryHandler(
        capacity=buffer_num_records,
        flushLevel=logging.ERROR,
        target=file_handler,
        flush_period=flush_period,
    )

    formatter = logging.Formatter(format, datefmt=datefmt)
    file_handler.setFormatter(formatter)
    root_logger.handlers.clear()
    root_logger.addHandler(mem_handler)

    # Trap SIGTERM, SIGINT to avoid broken pipes:
    SigHandler()
    multiprocessing_logging.install_mp_handler(logger=root_logger)
    root_logger.info(f"Configured logging on {socket.gethostname()}")
    sys.excepthook = log_uncaught_exceptions


def log_uncaught_exceptions(exctype: Any, value: Any, tb: Any) -> None:
    root_logger = logging.getLogger("balsam")
    root_logger.error(f"Uncaught Exception {exctype}: {value}", exc_info=(exctype, value, tb))


class Spinner(object):
    """
    https://github.com/click-contrib/click-spinner
    """

    spinner_cycle = itertools.cycle(["-", "/", "|", "\\"])

    def __init__(self, msg: str, stream: TextIO = sys.stdout) -> None:
        self.msg = msg
        self.stream = stream
        self.stop_running: Optional[threading.Event] = None
        self.spin_thread: Optional[threading.Thread] = None

    def start(self) -> None:
        self.stream.write(self.msg + "  ")
        if self.stream.isatty():
            self.stop_running = threading.Event()
            self.spin_thread = threading.Thread(target=self.init_spin)
            self.spin_thread.start()

    def stop(self) -> None:
        assert self.stop_running is not None
        if self.spin_thread:
            self.stop_running.set()
            self.spin_thread.join()

    def init_spin(self) -> None:
        assert self.stop_running is not None
        while not self.stop_running.is_set():
            self.stream.write(next(self.spinner_cycle))
            self.stream.flush()
            self.stop_running.wait(0.25)
            self.stream.write("\b")
            self.stream.flush()

    def __enter__(self) -> "Spinner":
        self.start()
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        self.stop()
        return None



import logging
import queue
from collections import defaultdict
from datetime import datetime
from typing import TYPE_CHECKING, Any, Dict, List, Optional, cast

from balsam.schemas import JobState
from balsam.util import Process, SigHandler

from .util import Queue

if TYPE_CHECKING:
    from balsam.client import RESTClient

logger = logging.getLogger(__name__)


class StatusUpdater(Process):
    def __init__(self, client: "RESTClient") -> None:
        super().__init__()
        self.client = client
        self.queue: "Queue[Dict[str, Any]]" = Queue()

    def _run(self) -> None:
        self.client.close_session()
        sig_handler = SigHandler()

        while not sig_handler.is_set():
            try:
                item = self.queue.get(block=True, timeout=1)
            except queue.Empty:
                continue

            updates = [item]
            while len(updates) < 10_000:
                try:
                    item = self.queue.get(block=True, timeout=1)
                    updates.append(item)
                except queue.Empty:
                    break
            self._perform_updates(updates)

        logger.info("Signal: break out of StatusUpdater main loop")
        self._drain_queue()
        logger.info("StatusUpdater thread finished.")

    def _drain_queue(self) -> None:
        updates = []
        while True:
            try:
                item = self.queue.get_nowait()
                updates.append(item)
            except queue.Empty:
                break
        if updates:
            self._perform_updates(updates)

    def put(
        self,
        id: int,
        state: JobState,
        state_timestamp: Optional[datetime] = None,
        state_data: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> None:
        if state_data is None:
            state_data = {}
        self.queue.put_nowait(
            {
                "id": id,
                "state": state,
                "state_timestamp": state_timestamp,
                "state_data": state_data,
                **kwargs,
            }
        )

    def _perform_updates(self, updates: List[Dict[str, Any]]) -> None:
        raise NotImplementedError


class BulkStatusUpdater(StatusUpdater):
    def _perform_updates(self, updates: List[Dict[str, Any]]) -> None:
        """
        In case two job updates occur in the same window,
        we sort updates by timestamp and avoid duplicate job
        updates in the same API call
        """
        updates_by_id = cast(Dict[Any, List[Dict[str, Any]]], defaultdict(list))
        for update in sorted(updates, key=lambda x: x["state_timestamp"]):  # type: ignore
            updates_by_id[update["id"]].append(update)

        while updates_by_id:
            bulk_update = [update_list.pop(0) for update_list in updates_by_id.values()]
            self.client.bulk_patch("jobs/", bulk_update)
            logger.info(f"StatusUpdater bulk-updated {len(bulk_update)} jobs")
            updates_by_id = {k: v for k, v in updates_by_id.items() if v}



import getpass
import logging
import os
import stat
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Dict, List, Type

from balsam.platform.scheduler import (
    SchedulerDeleteError,
    SchedulerError,
    SchedulerNonZeroReturnCode,
    SchedulerSubmitError,
)
from balsam.schemas import AllowedQueue, BatchJobState, SchedulerJobStatus
from balsam.site import ScriptTemplate

from .service_base import BalsamService

if TYPE_CHECKING:
    from balsam._api.models import BatchJob
    from balsam.client import RESTClient
    from balsam.platform.scheduler import SchedulerInterface  # noqa: F401

logger = logging.getLogger(__name__)


class SchedulerService(BalsamService):
    def __init__(
        self,
        client: "RESTClient",
        site_id: int,
        scheduler_class: Type["SchedulerInterface"],
        sync_period: int,
        allowed_queues: Dict[str, AllowedQueue],
        allowed_projects: List[str],
        optional_batch_job_params: Dict[str, str],
        job_template_path: Path,
        submit_directory: Path,
        filter_tags: Dict[str, str],
    ) -> None:
        super().__init__(client=client, service_period=sync_period)
        self.site_id = site_id
        self.scheduler = scheduler_class()
        self.allowed_queues = allowed_queues
        self.allowed_projects = allowed_projects
        self.optional_batch_job_params = optional_batch_job_params
        self.job_template = ScriptTemplate(job_template_path)
        self.submit_directory = submit_directory
        self.username = getpass.getuser()
        self.filter_tags = filter_tags

    def fail_submit(self, job: "BatchJob", msg: str) -> None:
        job.state = BatchJobState.submit_failed
        job.status_info = {**(job.status_info or {}), "error": msg}
        logger.error(f"Submit failed for BatchJob {job.id}: {msg}")

    def submit_launch(self, job: "BatchJob", scheduler_jobs: Dict[int, SchedulerJobStatus]) -> None:
        try:
            job.validate(
                self.allowed_queues,
                self.allowed_projects,
                self.optional_batch_job_params,
            )
        except ValueError as e:
            return self.fail_submit(job, str(e))

        num_queued = len([j for j in scheduler_jobs.values() if j.queue == job.queue])
        queue = self.allowed_queues[job.queue]
        if num_queued >= queue.max_queued_jobs:
            return self.fail_submit(job, f"Exceeded max {queue.max_queued_jobs} jobs in queue {job.queue}")

        script = self.job_template.render(
            project=job.project,
            queue=job.queue,
            num_nodes=job.num_nodes,
            wall_time_min=job.wall_time_min,
            job_mode=job.job_mode,
            filter_tags=job.filter_tags,
            partitions=job.partitions_to_cli_args(),
            **job.optional_params,
        )

        script_path = self.submit_directory.joinpath(f"qlaunch{job.id:05d}.sh")
        with open(script_path, "w") as fp:
            fp.write(script)

        logger.info(f"Generated script for batch_job {job.id} in {script_path}")

        st = os.stat(script_path)
        os.chmod(script_path, st.st_mode | stat.S_IEXEC)

        try:
            scheduler_id = self.scheduler.submit(
                script_path,
                project=job.project,
                queue=job.queue,
                num_nodes=job.num_nodes,
                wall_time_min=job.wall_time_min,
                cwd=self.submit_directory,
            )
        except SchedulerSubmitError as e:
            return self.fail_submit(job, f"Scheduler submit error:\n{e}")
        else:
            assert scheduler_id is not None
            job.scheduler_id = scheduler_id
            job.state = BatchJobState.queued
            job.status_info = {
                **(job.status_info or {}),
                "submit_script": script_path.as_posix(),
                "submit_time": str(datetime.utcnow()),
            }
            logger.info(f"Submit OK: {job}")

    def run_cycle(self) -> None:
        BatchJob = self.client.BatchJob
        api_jobs = list(
            BatchJob.objects.filter(
                site_id=self.site_id,
                state=["pending_submission", "queued", "running", "pending_deletion"],
            )
        )
        logger.debug(f"Fetched API BatchJobs: {[(j.id, j.state) for j in api_jobs]}")
        try:
            scheduler_jobs = self.scheduler.get_statuses(user=self.username)
        except SchedulerNonZeroReturnCode as exc:
            logger.error(f"scheduler.get_statuses nonzero return: {exc}")
            return

        for job in api_jobs:
            if job.state == "finished":
                continue
            if job.state == "pending_submission":
                if all(item in job.filter_tags.items() for item in self.filter_tags.items()):
                    self.submit_launch(job, scheduler_jobs)
                else:
                    logger.debug(
                        f"Will not submit batchjob {job.id}: does not match "
                        f"the current filter_tags criteria: {self.filter_tags}"
                    )
            elif job.state == "pending_deletion" and job.scheduler_id in scheduler_jobs:
                logger.info(f"Performing queue-deletion of batch job {job.id}")
                try:
                    self.scheduler.delete_job(job.scheduler_id)
                except SchedulerDeleteError as exc:
                    logger.warning(f"Failed to delete job {job.scheduler_id}: {exc}")
            elif job.scheduler_id not in scheduler_jobs:
                logger.info(
                    f"batch job {job.id}: scheduler_id {job.scheduler_id} no longer in queue statuses: finished"
                )
                job.state = BatchJobState.finished
                assert job.scheduler_id is not None
                assert job.status_info is not None
                job_log = self.scheduler.parse_logs(job.scheduler_id, job.status_info.get("submit_script", None))
                start_time = job_log.start_time
                end_time = job_log.end_time
                if start_time:
                    job.start_time = start_time
                if end_time:
                    job.end_time = end_time
            elif job.state != scheduler_jobs[job.scheduler_id].state:
                job.state = scheduler_jobs[job.scheduler_id].state
                logger.info(f"Job {job.id} (sched_id {job.scheduler_id}) advanced to state {job.state}")
        BatchJob.objects.bulk_update(api_jobs)
        self.update_site_info(scheduler_jobs)

    def update_site_info(self, scheduler_jobs: Dict[int, SchedulerJobStatus]) -> None:
        """Update on Site from nodelist & qstat"""
        # TODO: Periodically update Site nodelist; queues from here:
        site = self.client.Site.objects.get(id=self.site_id)
        try:
            site.backfill_windows = self.scheduler.get_backfill_windows()
        except SchedulerError as e:
            logger.warning(f"get_backfill_windows() error: {e}")
            logger.warning("Clearing site backfill windows; could not obtain")
            site.backfill_windows = {}
        site.queued_jobs = scheduler_jobs
        site.save()
        logger.debug(f"Updated Site info: {site.display_dict()}")

    def cleanup(self) -> None:
        logger.info("SchedulerService exiting")



import logging

from balsam._api import models
from balsam._api.app import ApplicationDefinition
from balsam._api.models import App as _APIApp, BatchJob, EventLog, Job, Session, Site, TransferItem
from balsam.config import ClientSettings, SiteConfig
from balsam.schemas import (
    AppParameter,
    BatchJobPartition,
    JobMode,
    JobState,
    JobTransferItem,
    TransferDirection,
    TransferSlot,
)

logger = logging.getLogger(__name__)

try:
    client = ClientSettings.load_from_file().build_client()
except Exception as exc:
    client = None  # type: ignore
    logger.warning(f"balsam.api failed to auto-load Client:\n{exc}")
else:
    Site.objects = models.SiteManager(client)
    _APIApp.objects = models.AppManager(client)
    BatchJob.objects = models.BatchJobManager(client)
    Job.objects = models.JobManager(client)
    TransferItem.objects = models.TransferItemManager(client)
    Session.objects = models.SessionManager(client)
    EventLog.objects = models.EventLogManager(client)
    ApplicationDefinition._set_client(client)
try:
    site_config = SiteConfig()
except Exception as exc:
    site_config = None  # type: ignore
    logger.debug(f"balsam.api failed to auto-load SiteConfig:\n{exc}")

__all__ = [
    "Site",
    "ApplicationDefinition",
    "BatchJob",
    "Job",
    "Session",
    "TransferItem",
    "EventLog",
    "JobTransferItem",
    "JobState",
    "AppParameter",
    "TransferDirection",
    "TransferSlot",
    "JobMode",
    "BatchJobPartition",
    "site_config",
]



from datetime import datetime
from logging import getLogger
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union, cast

from fastapi import HTTPException, status
from fastapi.encoders import jsonable_encoder
from sqlalchemy import case, func, literal_column, orm
from sqlalchemy.orm import Query, Session

from balsam import schemas
from balsam.schemas.job import JobState
from balsam.server import ValidationError, models
from balsam.server.routers.filters import JobQuery
from balsam.server.utils import Paginator

logger = getLogger(__name__)


def owned_job_query(
    db: Session, owner: schemas.UserOut, with_parents: bool = False, id_only: bool = False
) -> "Query[models.Job]":
    qs: "Query[models.Job]" = db.query(models.Job.id) if id_only else db.query(models.Job)
    if with_parents:
        qs = qs.options(orm.selectinload(models.Job.parents).load_only(models.Job.id))  # type: ignore
    qs = qs.join(models.App).join(models.Site).filter(models.Site.owner_id == owner.id)  # type: ignore
    return qs


def set_parent_ids(jobs: Iterable[models.Job]) -> None:
    for job in jobs:
        job.parent_ids = [p.id for p in cast(Iterable[models.Job], job.parents)]


def update_states_by_query(
    qs: "Query[models.Job]",
    state: str,
    data: Union[Dict[str, Any], str, None] = None,
    timestamp: Optional[datetime] = None,
) -> Tuple[List[models.Job], List[models.LogEvent]]:
    now = datetime.utcnow()
    if isinstance(data, str):
        data = {"message": data}
    elif data is None:
        data = {}

    updated_jobs = qs.all()
    events = []
    for job in updated_jobs:
        event = _update_state(job, state, now, data)
        if event:
            events.append(event)
    return updated_jobs, events


def populate_transfers(db_job: models.Job, job_spec: schemas.ServerJobCreate) -> List[models.TransferItem]:
    for transfer_name, transfer_spec in job_spec.transfers.items():
        transfer_slot = db_job.app.transfers.get(transfer_name)
        if transfer_slot is None:
            raise ValidationError(f"App {db_job.app.name} has no Transfer slot named {transfer_name}")
        direction, local_path = transfer_slot["direction"], transfer_slot["local_path"]
        recursive = transfer_slot["recursive"]
        assert direction in ["in", "out"]
        location_alias, remote_path = (
            transfer_spec.location_alias,
            transfer_spec.path.as_posix(),
        )
        url = db_job.app.site.transfer_locations.get(location_alias)
        if url is None:
            raise ValidationError(f"Site has no Transfer URL named {location_alias}")

        transfer_item = models.TransferItem(
            direction=direction,
            local_path=local_path,
            remote_path=remote_path,
            recursive=recursive,
            location_alias=location_alias,
            state="pending" if direction == "in" else "awaiting_job",
            task_id="",
            transfer_info={},
        )
        db_job.transfer_items.append(transfer_item)
    return db_job.transfer_items


def fetch(
    db: Session,
    owner: schemas.UserOut,
    paginator: Optional[Paginator[models.Job]] = None,
    job_id: Optional[int] = None,
    filterset: Optional[JobQuery] = None,
) -> "Tuple[int, Union[List[models.Job], Query[models.Job]]]":
    qs = owned_job_query(db, owner, with_parents=True)
    if job_id is not None:
        qs = qs.filter(models.Job.id == job_id)
    if filterset:
        qs = filterset.apply_filters(qs)
    if paginator is None:
        job = qs.one()
        set_parent_ids([job])
        return 1, [job]
    count = qs.group_by(models.Job.id).count()
    jobs = paginator.paginate(qs)
    set_parent_ids(jobs)
    return count, jobs


def bulk_create(
    db: Session, owner: schemas.UserOut, job_specs: List[schemas.ServerJobCreate]
) -> Tuple[List[models.Job], List[models.LogEvent], List[models.TransferItem]]:
    now = datetime.utcnow()
    app_ids = set(job.app_id for job in job_specs)
    parent_ids = set(pid for job in job_specs for pid in job.parent_ids)

    apps: Dict[int, models.App] = {
        app.id: app
        for app in db.query(models.App)
        .join(models.App.site)  # type: ignore
        .filter(models.Site.owner_id == owner.id, models.App.id.in_(app_ids))
        .options(orm.selectinload(models.App.site).load_only(models.Site.transfer_locations))
        .all()
    }
    if len(apps) < len(app_ids):
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail="Could not find one or more Apps. Double check app_id fields.",
        )

    parents = {job.id: job for job in owned_job_query(db, owner).filter(models.Job.id.in_(parent_ids)).all()}
    if len(parents) < len(parent_ids):
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail="Could not find one or more Jobs set as parents. Double check parent_ids fields.",
        )

    created_jobs, created_transfers, created_events = [], [], []
    for job_spec in job_specs:
        db_job = models.Job(**jsonable_encoder(job_spec.dict(exclude={"parent_ids", "transfers"})))
        db.add(db_job)
        db_job.app = apps[job_spec.app_id]

        assert isinstance(db_job.parents, list)
        db_job.parents.extend(parents[id] for id in job_spec.parent_ids)
        populate_transfers(db_job, job_spec)
        if any(job.state != JobState.job_finished for job in db_job.parents):
            db_job.state = JobState.awaiting_parents
        else:
            if any(tr.direction == "in" for tr in db_job.transfer_items) > 0:
                db_job.state = JobState.ready
            else:
                db_job.state = JobState.staged_in

        event = models.LogEvent(
            job=db_job,
            timestamp=now,
            from_state=JobState.created,
            to_state=db_job.state,
        )
        created_jobs.append(db_job)
        created_events.append(event)
        created_transfers.extend(db_job.transfer_items)

    set_parent_ids(created_jobs)
    db.flush()
    logger.debug(f"Bulk-created {len(created_jobs)} jobs")
    return created_jobs, created_events, created_transfers


def _update_state(
    job: models.Job, state: str, state_timestamp: datetime, state_data: Dict[str, Any]
) -> Optional[models.LogEvent]:
    if state == job.state or state is None:
        return None
    if state == "RESET" and job.state in ["AWAITING_PARENTS", "READY"]:
        return None
    event = models.LogEvent(
        job=job,
        from_state=job.state,
        to_state=state,
        timestamp=state_timestamp,
        data=state_data,
    )
    assert isinstance(event.data, dict)
    job.state = state

    if job.state != "RUNNING" and job.session_id is not None:
        job.session = None
        job.session_id = None

    if job.state == "RESET":
        if all(parent.state == "JOB_FINISHED" for parent in cast(Iterable[models.Job], job.parents)):
            job.state = "READY"
        else:
            job.state = "AWAITING_PARENTS"

    if job.state == "READY":
        if all(item.state == "done" for item in job.transfer_items if item.direction == "in"):
            job.state = "STAGED_IN"
            event.data["message"] = "Skipped stage in"

    if job.state == "POSTPROCESSED":
        if all(item.state == "done" for item in job.transfer_items if item.direction == "out"):
            job.state = "JOB_FINISHED"
            event.data["message"] = "Skipped stage out"
        else:
            for item in job.transfer_items:
                if item.state == "awaiting_job":
                    item.state = "pending"

    if job.state == "STAGED_OUT":
        job.state = "JOB_FINISHED"

    event.to_state = job.state
    return event


def _update(job: models.Job, data: Dict[str, Any]) -> Optional[models.LogEvent]:
    state = data.pop("state", None)
    state_timestamp = data.pop("state_timestamp", datetime.utcnow())
    state_data = data.pop("state_data", {})

    workdir = data.pop("workdir", None)
    if workdir:
        job.workdir = str(workdir)
    for k, v in data.items():
        setattr(job, k, v)
    job.return_code = data.get("return_code", job.return_code)
    job.data = data.get("data", job.data)
    event = _update_state(job, state, state_timestamp, state_data)
    return event


def _check_waiting_children(db: Session, parent_ids: Iterable[int]) -> Tuple[List[models.Job], List[models.LogEvent]]:
    parent_alias1 = orm.aliased(models.Job)
    parent_alias2 = orm.aliased(models.Job)
    num_parents = func.count(parent_alias1.id)
    num_finished_parents = func.count(
        case(
            [((parent_alias1.state == "JOB_FINISHED"), parent_alias1.id)],
            else_=literal_column("NULL"),
        )
    )
    ready_children: List[models.Job] = (
        db.query(models.Job)
        .options(  # type: ignore
            orm.selectinload(models.Job.transfer_items).load_only(
                models.TransferItem.state,
                models.TransferItem.direction,
            ),
            # TODO: what is most efficient way to load parent IDs in same query?
            orm.joinedload(models.Job.parents).load_only(
                models.Job.id,
                models.Job.state,
            ),
        )
        .join(models.Job.parents.of_type(parent_alias1))
        .join(models.Job.parents.of_type(parent_alias2))
        .filter(models.Job.state == "AWAITING_PARENTS")
        .filter(parent_alias2.id.in_(parent_ids))
        .group_by(models.Job.id)
        .having(num_parents == num_finished_parents)
        .all()
    )
    now = datetime.utcnow()
    new_events = []
    for child in ready_children:
        event = _update_state(child, "READY", now, {"message": "All parents finished"})
        if event:
            new_events.append(event)
    return ready_children, new_events


def _select_jobs_for_update(qs: "Query[models.Job]") -> List[models.Job]:
    qs = qs.options(  # type: ignore
        orm.selectinload(models.Job.parents).load_only(
            models.Job.id,
            models.Job.state,
        ),
        orm.selectinload(models.Job.transfer_items).load_only(
            models.TransferItem.state,
            models.TransferItem.direction,
        ),
    ).with_for_update(of=models.Job)
    return qs.all()


def _update_jobs(
    db: Session, update_jobs: List[models.Job], patch_dicts: Dict[int, Dict[str, Any]]
) -> Tuple[List[models.Job], List[models.LogEvent]]:
    new_events = []
    # First, perform job-wise updates:
    for job in update_jobs:
        event = _update(job, patch_dicts[job.id])
        if event:
            new_events.append(event)
    update_ids = [job.id for job in update_jobs]
    db.flush()

    # Then update affected children in a second query:
    updated_children, child_events = _check_waiting_children(db, update_ids)
    db.flush()
    update_jobs.extend(updated_children)
    new_events.extend(child_events)
    set_parent_ids(update_jobs)
    return update_jobs, new_events


def bulk_update(
    db: Session, owner: schemas.UserOut, patch_dicts: Dict[int, Dict[str, Any]]
) -> Tuple[List[models.Job], List[models.LogEvent]]:
    job_ids = set(patch_dicts.keys())
    qs = owned_job_query(db, owner).filter(models.Job.id.in_(job_ids))
    update_jobs = _select_jobs_for_update(qs)
    if len(update_jobs) < len(patch_dicts):
        raise ValidationError("Could not find some Job IDs")
    return _update_jobs(db, update_jobs, patch_dicts)


def update_query(
    db: Session, owner: schemas.UserOut, update_data: Dict[str, Any], filterset: JobQuery
) -> Tuple[List[models.Job], List[models.LogEvent]]:
    qs = owned_job_query(db, owner)
    qs = filterset.apply_filters(qs)
    update_jobs = _select_jobs_for_update(qs)
    patch_dicts = {job.id: update_data.copy() for job in update_jobs}
    return _update_jobs(db, update_jobs, patch_dicts)


def delete_query(
    db: Session,
    owner: schemas.UserOut,
    filterset: Optional[JobQuery] = None,
    job_id: Optional[int] = None,
) -> int:
    qs = owned_job_query(db, owner, id_only=True)
    if job_id is not None:
        qs = qs.filter(models.Job.id == job_id)
        qs.one()
    else:
        assert filterset is not None
        qs = filterset.apply_filters(qs)
    qs = qs.filter(models.Job.session_id.is_(None)).with_for_update(of=models.Job, skip_locked=True)  # type: ignore
    num_deleted = db.query(models.Job).filter(models.Job.id.in_(qs)).delete(synchronize_session=False)
    db.flush()
    logger.debug(f"Deleted {num_deleted} jobs")
    return num_deleted



from typing import Generic, Optional, TypeVar

from fastapi import Query as APIQuery
from sqlalchemy.orm import Query as SQLQuery

T = TypeVar("T")


class Paginator(Generic[T]):
    def __init__(
        self,
        limit: Optional[int] = APIQuery(None, description="Maximum number of items to return."),
        offset: int = APIQuery(0, description="Starting index from which to retrieve results."),
    ):
        self.limit = limit
        self.offset = offset

    def paginate(self, iterable: "SQLQuery[T]") -> "SQLQuery[T]":
        if self.limit is not None:
            return iterable[self.offset : self.offset + self.limit]  # type: ignore
        return iterable[self.offset :]  # type:ignore

    def __new__(
        cls,
        limit: Optional[int] = APIQuery(None, description="Maximum number of items to return."),
        offset: int = APIQuery(0, description="Starting index from which to list items."),
    ) -> "Paginator[T]":
        return super().__new__(cls)  # type: ignore



import multiprocessing
from multiprocessing.queues import Queue as QueueBase
from typing import TYPE_CHECKING, Any, Generic, TypeVar, cast

# The following implementation of custom MyQueue to avoid NotImplementedError
# when calling queue.qsize() in MacOS X comes almost entirely from this github
# discussion: https://github.com/keras-team/autokeras/issues/368
# Necessary modification is made to make the code compatible with Python3.


class SharedCounter(object):
    """A synchronized shared counter.
    The locking done by multiprocessing.Value ensures that only a single
    process or thread may read or write the in-memory ctypes object. However,
    in order to do n += 1, Python performs a read followed by a write, so a
    second process may read the old value before the new one is written by the
    first process. The solution is to use a multiprocessing.Lock to guarantee
    the atomicity of the modifications to Value.
    This class comes almost entirely from Eli Bendersky's blog:
    http://eli.thegreenplace.net/2012/01/04/shared-counter-with-pythons-multiprocessing/
    """

    def __init__(self, n: int = 0) -> None:
        self.count = multiprocessing.Value("i", n)

    def increment(self, n: int = 1) -> None:
        """Increment the counter by n (default = 1)"""
        with self.count.get_lock():
            self.count.value += n

    @property
    def value(self) -> int:
        """Return the value of the counter"""
        return cast(int, self.count.value)


T = TypeVar("T")
if TYPE_CHECKING:

    class _QueueBase(QueueBase[T]):
        pass


else:

    class _QueueBase(Generic[T], QueueBase):
        pass


class _FallbackQueue(_QueueBase[T]):
    """A portable implementation of multiprocessing.Queue.
    Because of multithreading / multiprocessing semantics, Queue.qsize() may
    raise the NotImplementedError exception on Unix platforms like Mac OS X
    where sem_getvalue() is not implemented. This subclass addresses this
    problem by using a synchronized shared counter (initialized to zero) and
    increasing / decreasing its value every time the put() and get() methods
    are called, respectively. This not only prevents NotImplementedError from
    being raised, but also allows us to implement a reliable version of both
    qsize() and empty().
    """

    def __init__(self) -> None:
        super().__init__(ctx=multiprocessing.get_context())
        self.size = SharedCounter(0)

    def put(self, *args: Any, **kwargs: Any) -> None:
        super().put(*args, **kwargs)
        self.size.increment(1)

    def get(self, *args: Any, **kwargs: Any) -> T:
        result = super().get(*args, **kwargs)
        self.size.increment(-1)
        return result

    def qsize(self) -> int:
        """Reliable implementation of multiprocessing.Queue.qsize()"""
        return self.size.value

    def empty(self) -> bool:
        """Reliable implementation of multiprocessing.Queue.empty()"""
        return not self.qsize()


try:
    multiprocessing.Queue().qsize()
except NotImplementedError:
    Queue = _FallbackQueue
else:
    Queue = multiprocessing.Queue  # type: ignore



import logging
import math
from datetime import datetime, timedelta
from typing import List, Optional, Tuple

from sqlalchemy import orm

from balsam import schemas
from balsam.server import ValidationError, models
from balsam.server.routers.filters import SessionQuery

from .jobs import set_parent_ids, update_states_by_query

logger = logging.getLogger(__name__)

SESSION_EXPIRE_PERIOD = timedelta(minutes=5)
SESSION_SWEEP_PERIOD = timedelta(minutes=3)
_sweep_time: Optional[datetime] = None
Session = orm.Session
Query = orm.Query


def owned_session_query(db: Session, owner: schemas.UserOut) -> "Query[models.Session]":
    return db.query(models.Session).join(models.Site).filter(models.Site.owner_id == owner.id)  # type: ignore


def fetch(db: Session, owner: schemas.UserOut, filterset: SessionQuery) -> Tuple[int, List[models.Session]]:
    qs = owned_session_query(db, owner)
    result = filterset.apply_filters(qs).all()
    count = len(result)
    return count, result


def _clear_stale_sessions(db: Session, owner: schemas.UserOut) -> Tuple[List[models.Job], List[models.LogEvent]]:
    global _sweep_time
    now = datetime.utcnow()
    if _sweep_time is not None and now < _sweep_time + SESSION_SWEEP_PERIOD:
        return [], []
    _sweep_time = now
    expiry_time = now - SESSION_EXPIRE_PERIOD
    expired_sessions = owned_session_query(db, owner).filter(models.Session.heartbeat <= expiry_time).all()

    expired_jobs, expiry_events = [], []
    for session in expired_sessions:
        zombies = session.jobs.filter(models.Job.state == "RUNNING")
        jobs, events = update_states_by_query(
            zombies,
            state="RUN_TIMEOUT",
            data="Session expired: lost contact with launcher",
        )
        set_parent_ids(jobs)
        expired_jobs.extend(jobs)
        expiry_events.extend(events)
    sess_ids = [sess.id for sess in expired_sessions]
    db.query(models.Session).filter(models.Session.id.in_(sess_ids)).delete(synchronize_session=False)
    db.flush()
    if expired_sessions:
        logger.info(f"Deleting stale sessions: {[s.id for s in expired_sessions]}")
    return expired_jobs, expiry_events


def create(
    db: Session, owner: schemas.UserOut, session: schemas.SessionCreate
) -> Tuple[models.Session, List[models.Job], List[models.LogEvent]]:
    expired_jobs, expiry_events = _clear_stale_sessions(db, owner)

    site_id = (
        db.query(models.Site.id).filter(models.Site.owner_id == owner.id, models.Site.id == session.site_id).scalar()  # type: ignore
    )
    if site_id is None:
        raise ValidationError(f"No site with ID {session.site_id}")
    if session.batch_job_id is not None:
        batch_job_id = (
            db.query(models.BatchJob.id)
            .filter(models.BatchJob.site_id == site_id)
            .filter(models.BatchJob.id == session.batch_job_id)
        ).scalar()  # type: ignore
        if batch_job_id is None:
            raise ValidationError(f"No batch_job with id {session.batch_job_id}")

    created_session = models.Session(**session.dict())
    db.add(created_session)
    db.flush()
    return created_session, expired_jobs, expiry_events


def acquire(
    db: Session, owner: schemas.UserOut, session_id: int, spec: schemas.SessionAcquire
) -> Tuple[List[models.Job], List[models.Job], List[models.LogEvent]]:
    expired_jobs, expiry_events = _clear_stale_sessions(db, owner)
    session = (owned_session_query(db, owner).filter(models.Session.id == session_id)).one()
    session.heartbeat = datetime.utcnow()

    qs = db.query(models.Job).join(models.App)  # type: ignore
    qs = qs.options(orm.selectinload(models.Job.parents).load_only(models.Job.id))
    qs = qs.filter(models.App.site_id == session.site_id)  # At site
    qs = qs.filter(models.Job.session_id.is_(None))  # type: ignore # Unlocked

    if spec.app_ids:
        qs = qs.filter(models.Job.app_id.in_(spec.app_ids))

    qs = qs.filter(models.Job.state.in_(spec.states))  # Matching states
    logger.debug(f"Acquire: filtering for jobs with states: {spec.states}")

    qs = qs.filter(models.Job.tags.contains(spec.filter_tags))  # type: ignore # Matching tags
    if spec.max_wall_time_min:
        qs = qs.filter(models.Job.wall_time_min <= spec.max_wall_time_min)  # By time

    if spec.max_nodes_per_job:
        qs = qs.filter(models.Job.num_nodes <= spec.max_nodes_per_job)
    if spec.serial_only:
        qs = qs.filter(models.Job.num_nodes == 1, models.Job.ranks_per_node == 1)
    qs = qs.with_for_update(of=models.Job, skip_locked=True)
    qs = qs.order_by(
        models.Job.num_nodes.desc(),
        models.Job.node_packing_count,
        models.Job.wall_time_min.desc(),
    )

    if spec.max_aggregate_nodes is not None:
        aggregate_nodes = spec.max_aggregate_nodes + 0.001
    else:
        aggregate_nodes = math.inf
    jobs = list(qs[: spec.max_num_jobs])
    idx = 0
    acquired_jobs = []

    while idx < len(jobs) and aggregate_nodes > 0.002:
        job = jobs[idx]
        job_footprint = job.num_nodes / job.node_packing_count
        if job_footprint <= aggregate_nodes:
            acquired_jobs.append(job)
            aggregate_nodes -= job_footprint
            idx += 1
        else:
            idx = next(
                (
                    i
                    for (i, job) in enumerate(jobs[idx:], idx)
                    if job.num_nodes / job.node_packing_count <= aggregate_nodes
                ),
                len(jobs),
            )

    for job in acquired_jobs:
        job.session_id = session.id
        # Do not overwrite job.batch_job_id with a Session that has no batch_job_id:
        if session.batch_job_id is not None:
            job.batch_job_id = session.batch_job_id
        job.parent_ids = [parent.id for parent in job.parents]
    db.flush()
    logger.debug(f"Acquired {len(acquired_jobs)} jobs")
    return acquired_jobs, expired_jobs, expiry_events


def tick(
    db: Session, owner: schemas.UserOut, session_id: int
) -> Tuple[datetime, List[models.Job], List[models.LogEvent]]:
    expired_jobs, expiry_events = _clear_stale_sessions(db, owner)
    in_db = owned_session_query(db, owner).filter(models.Session.id == session_id).one()
    ts = datetime.utcnow()
    in_db.heartbeat = ts
    db.flush()
    return ts, expired_jobs, expiry_events


def delete(db: Session, owner: schemas.UserOut, session_id: int) -> None:
    qs = owned_session_query(db, owner).filter(models.Session.id == session_id)
    qs.one()
    db.query(models.Session).filter(models.Session.id == session_id).delete()
    db.flush()



import datetime
import logging
import os
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Union

from dateutil.parser import ParserError

from balsam.util import parse_to_utc

from .scheduler import (
    SchedulerBackfillWindow,
    SchedulerJobLog,
    SchedulerJobStatus,
    SubprocessSchedulerInterface,
    scheduler_subproc,
)

PathLike = Union[Path, str]
logger = logging.getLogger(__name__)


# parse "00:00:00" to minutes
def parse_clock(t_str: str) -> int:
    parts = t_str.split(":")
    n = len(parts)
    H = M = S = 0
    if n == 3:
        H, M, S = map(int, parts)
    elif n == 2:
        M, S = map(int, parts)

    return H * 60 + M + round(S / 60)


# parse "1-00:00:00" to minutes
def parse_time_minutes(t_str: str) -> Optional[int]:
    mins = 0
    try:
        parts = t_str.split("-")
        if len(parts) == 1:
            mins += parse_clock(parts[0])
        elif len(parts) == 2:
            mins += parse_clock(parts[1])
            mins += int(parts[0]) * 24 * 60
    except ValueError:
        return None
    else:
        return mins


# parse "2021-02-20T13:11:53" to minutes
def parse_queued_time(t_str: str) -> Optional[int]:
    mins = 0.0
    try:
        submit_time = datetime.datetime.strptime(t_str, "%Y-%m-%dT%H:%M:%S")
        timediff = datetime.datetime.now() - submit_time
        mins = timediff.total_seconds() / 60.0
    except Exception:
        logger.exception(f"exception while processing {t_str}")
        return None
    else:
        return int(mins)


class SlurmScheduler(SubprocessSchedulerInterface):
    status_exe = "squeue"
    submit_exe = "sbatch"
    delete_exe = "scancel"
    backfill_exe = "sinfo"
    history_exe = "sacct"
    default_submit_kwargs: Dict[str, str] = {}
    submit_kwargs_flag_map: Dict[str, str] = {}

    # maps scheduler states to Balsam states
    job_states = {
        "PENDING": "queued",
        "CONFIGURING": "running",
        "RUNNING": "running",
        "COMPLETING": "running",
        "RESV_DEL_HOLD": "queued",
        "ADMIN_HOLD": "queued",
        "FINISHED": "finished",
        "FAILED": "failed",
        "CANCELLED": "finished",
        "DEADLINE": "finished",
        "PREEMPTED": "finished",
        "REQUEUED": "queued",
        "SIGNALING": "running",
        "STAGE_OUT": "running",
        "TIMEOUT": "failed",
    }

    @staticmethod
    def _job_state_map(scheduler_state: str) -> str:
        return SlurmScheduler.job_states.get(scheduler_state, "unknown")

    # maps Balsam status fields to the scheduler fields
    # should be a comprehensive list of scheduler status fields
    _status_fields = {
        "scheduler_id": "jobid",
        "state": "state",
        "queue": "partition",
        "num_nodes": "numnodes",
        "wall_time_min": "timelimit",
        "project": "account",
        "time_remaining_min": "timeleft",
        "queued_time_min": "submittime",
    }

    # when reading these fields from the scheduler apply
    # these maps to the string extracted from the output
    @staticmethod
    def _status_field_map(balsam_field: str) -> Optional[Callable[[str], Any]]:
        status_field_map = {
            "scheduler_id": lambda id: int(id),
            "state": SlurmScheduler._job_state_map,
            "queue": lambda queue: str(queue),
            "num_nodes": lambda n: int(n),
            "wall_time_min": parse_time_minutes,
            "project": lambda project: str(project),
            "time_remaining_min": parse_time_minutes,
            "queued_time_min": parse_queued_time,
        }
        return status_field_map.get(balsam_field, None)

    # maps node list states to Balsam node states
    # descriptions: https://slurm.schedmd.com/sinfo.html
    _node_states = {
        "alloc": "busy",  # allocated
        "boot": "busy",
        "comp": "busy",  # completing
        "down": "busy",
        "drain": "busy",  # drained
        "drng": "busy",  # draining
        "fail": "busy",
        "failg": "busy",  # failing
        "futr": "busy",  # future
        "idle": "idle",
        "maint": "busy",  # maintenance
        "mix": "busy",
        "npc": "busy",  # perfctrs
        "pow_dn": "busy",  # power down
        "pow_up": "busy",  # power up
        "resv": "busy",  # reserved
        "unk": "busy",  # unknown
    }

    @staticmethod
    def _node_state_map(nodelist_state: str) -> str:
        try:
            # removing special symbols that have some meaning
            # in the future we might want to encode this info
            # * The node is presently not responding and will not be allocated any new work.
            nodelist_state = nodelist_state.replace("*", "")
            # ~ The node is presently in a power saving mode
            nodelist_state = nodelist_state.replace("~", "")
            # # The node is presently being powered up or configured.
            nodelist_state = nodelist_state.replace("#", "")
            # % The node is presently being powered down.
            nodelist_state = nodelist_state.replace("%", "")
            # $ The node is currently in a reservation with a flag value of "maintenance".
            nodelist_state = nodelist_state.replace("$", "")
            # @ The node is pending reboot.
            nodelist_state = nodelist_state.replace("@", "")
            # alloc+ The node is allocated to one or more active jobs plus one or more jobs are in the process of COMPLETING.
            nodelist_state = nodelist_state.replace("+", "")
            return SlurmScheduler._node_states[nodelist_state]
        except KeyError:
            logger.warning("node state %s is not recognized", nodelist_state)
            return "unknown"

    # maps the Balsam status fields to the node list fields
    # should be a comprehensive list of node list fields
    _nodelist_fields = {
        "id": "nodelist",
        "queues": "partition",
        "node_state": "stateshort",
    }

    _fields_encondings = {
        "nodelist": "%N",
        "partition": "%P",
        "stateshort": "%t",
    }

    # when reading these fields from the scheduler apply
    # these maps to the string extracted from the output
    @staticmethod
    def _backfill_field_map(balsam_field: str) -> Callable[[str], Any]:
        nodelist_field_map = {
            "queues": lambda q: q.split(":"),
            "state": SlurmScheduler._node_state_map,
            "backfill_time": parse_time_minutes,
        }
        return nodelist_field_map.get(balsam_field, lambda x: x)

    @staticmethod
    def _set_envs() -> Dict[str, str]:
        env = {}
        fields = SlurmScheduler._status_fields.values()
        env["SQUEUE_FORMAT2"] = ",".join(fields)
        fields = SlurmScheduler._nodelist_fields.values()
        env["SINFO_FORMAT"] = " ".join(SlurmScheduler._fields_encondings[field] for field in fields)
        os.environ.update(env)
        return env

    @staticmethod
    def _render_submit_args(
        script_path: PathLike, project: str, queue: str, num_nodes: int, wall_time_min: int, **kwargs: Any
    ) -> List[str]:
        args = [
            SlurmScheduler.submit_exe,
            "-o",
            os.path.basename(os.path.splitext(script_path)[0]) + ".output",
            "-e",
            os.path.basename(os.path.splitext(script_path)[0]) + ".error",
            "-A",
            project,
            "-q",
            queue,
            "-N",
            str(int(num_nodes)),
            "-t",
            str(int(wall_time_min)),
        ]
        # adding additional flags as needed, e.g. `-C knl`
        for key, default_value in SlurmScheduler.default_submit_kwargs.items():
            flag = SlurmScheduler.submit_kwargs_flag_map[key]
            value = kwargs.setdefault(key, default_value)
            args += [flag, value]

        args.append(str(script_path))
        return args

    @staticmethod
    def _render_status_args(
        project: Optional[str] = None, user: Optional[str] = None, queue: Optional[str] = None
    ) -> List[str]:
        SlurmScheduler._set_envs()
        args = [SlurmScheduler.status_exe]
        if user is not None:
            args += ["-u", user]
        if project is not None:
            args += ["-A", project]
        if queue is not None:
            args += ["-q", queue]
        return args

    @staticmethod
    def _render_delete_args(job_id: Union[int, str]) -> List[str]:
        return [SlurmScheduler.delete_exe, str(job_id)]

    @staticmethod
    def _render_backfill_args() -> List[str]:
        SlurmScheduler._set_envs()
        return [SlurmScheduler.backfill_exe]

    @staticmethod
    def _parse_submit_output(submit_output: str) -> int:
        try:
            scheduler_id = int(submit_output)
        except ValueError:
            scheduler_id = int(submit_output.split()[-1])
        return scheduler_id

    @staticmethod
    def _parse_status_output(raw_output: str) -> Dict[int, SchedulerJobStatus]:
        # TODO: this can be much more efficient with a compiled regex findall()
        status_dict = {}
        job_lines = raw_output.strip().split("\n")[1:]
        for line in job_lines:
            try:
                job_stat = SlurmScheduler._parse_status_line(line)
            except (ValueError, TypeError):
                logger.exception(f"Cannot parse job status: {line}")
                continue
            else:
                status_dict[job_stat.scheduler_id] = job_stat
        return status_dict

    @staticmethod
    def _parse_status_line(line: str) -> SchedulerJobStatus:
        fields = line.split()
        actual = len(fields)
        expected = len(SlurmScheduler._status_fields)
        if actual != expected:
            logger.warning(f"Line has {actual} columns: expected {expected}:\n{fields}")
            raise ValueError(f"Line has {actual} columns: expected {expected}:\n{fields}")
        logger.debug(f"parsing fields: {fields}")
        status = {}
        for name, value in zip(SlurmScheduler._status_fields, fields):
            func = SlurmScheduler._status_field_map(name)
            if callable(func):
                status[name] = func(value)

        return SchedulerJobStatus(**status)

    @staticmethod
    def _parse_backfill_output(stdout: str) -> Dict[str, List[SchedulerBackfillWindow]]:
        # TODO: NERSC currently does not provide this kind of information
        return {}

    @staticmethod
    def _parse_logs(scheduler_id: Union[int, str], job_script_path: Optional[PathLike]) -> SchedulerJobLog:
        # use command to get batch job runtime
        args = [SlurmScheduler.history_exe]
        args += ["-o", "JobId,Start,End"]  # output format
        args += ["-j", str(scheduler_id)]  # for this job id
        args += ["-n"]  # no header
        stdout = scheduler_subproc(args)
        # ---- output like this:
        # 40589507     2021-03-12T08:40:12 2021-03-12T08:41:18
        # 40589507.ba+ 2021-03-12T08:40:12 2021-03-12T08:41:18
        # 40589507.ex+ 2021-03-12T08:40:12 2021-03-12T08:41:21
        # 40589507.0   2021-03-12T08:40:42 2021-03-12T08:40:45
        # returns empty string with newline for not found jobs:
        if len(stdout) <= 5:
            return SchedulerJobLog()
        first_line = stdout.strip().split("\n")[0]  # only take first line
        job_data = first_line.split()
        try:
            start_time = parse_to_utc(job_data[1])
            end_time = parse_to_utc(job_data[2])
        except ParserError:
            logger.warning(f"Failed to parse job_data times (START_TIME: {job_data[1]}) (FINISH_TIME: {job_data[2]})")
            return SchedulerJobLog()
        return SchedulerJobLog(start_time=start_time, end_time=end_time)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import setuptools
import re
from setuptools.command import easy_install

"""
Monkey patch setuptools to write faster console_scripts with this format:
    import sys
    from mymodule import entry_function
    sys.exit(entry_function())
This is better.
(c) 2016, Aaron Christianson
http://github.com/ninjaaron/fast-entry_points
"""

TEMPLATE = r"""
# -*- coding: utf-8 -*-
# EASY-INSTALL-ENTRY-SCRIPT: '{3}','{4}','{5}'
__requires__ = '{3}'
import re
import sys
from {0} import {1}
if __name__ == '__main__':
    sys.argv[0] = re.sub(r'(-script\.pyw?|\.exe)?$', '', sys.argv[0])
    sys.exit({2}())
""".lstrip()


@classmethod
def get_args(cls, dist, header=None):  # noqa: D205,D400
    """
    Yield write_script() argument tuples for a distribution's
    console_scripts and gui_scripts entry points.
    """
    if header is None:
        # pylint: disable=E1101
        header = cls.get_header()
    spec = str(dist.as_requirement())
    for type_ in "console", "gui":
        group = type_ + "_scripts"
        for name, ep in dist.get_entry_map(group).items():
            # ensure_safe_name
            if re.search(r"[\\/]", name):
                raise ValueError("Path separators not allowed in script names")
            script_text = TEMPLATE.format(ep.module_name, ep.attrs[0], ".".join(ep.attrs), spec, group, name)
            # pylint: disable=E1101
            args = cls._get_script_args(type_, name, header, script_text)
            for res in args:
                yield res


# pylint: disable=E1101
easy_install.ScriptWriter.get_args = get_args

setuptools.setup()



import logging
from datetime import datetime, timedelta
from typing import Any, Dict, Optional

import click
import dateutil.parser

from . import urls
from .requests_client import RequestsClient
from .rest_base_client import AuthError

logger = logging.getLogger(__name__)


class BasicAuthRequestsClient(RequestsClient):
    def __init__(
        self,
        api_root: str,
        username: Optional[str] = None,
        password: Optional[str] = None,
        token: Optional[str] = None,
        token_expiry: Optional[datetime] = None,
        connect_timeout: float = 6.2,
        read_timeout: float = 120.0,
        retry_count: int = 10,
    ) -> None:
        super().__init__(
            api_root,
            connect_timeout=connect_timeout,
            read_timeout=read_timeout,
            retry_count=retry_count,
        )
        self.username = username
        self.password = password
        self.token = token
        self.token_expiry = token_expiry
        self.expires_in = timedelta(hours=240)
        if self.token is not None:
            self.session.auth = None
            self.session.headers["Authorization"] = f"Bearer {self.token}"
            self._authenticated = True
        if self.token_expiry:
            self.expires_in = self.token_expiry - datetime.utcnow()
            if self.expires_in.total_seconds() <= 1:
                self._authenticated = False
                logger.warning("Auth Token is expired; please refresh with `balsam login`")
            elif self.expires_in < timedelta(hours=24):
                logger.warning(f"Auth Token will expire in {self.expires_in}; please refresh with `balsam login`")

    def refresh_auth(self) -> None:
        if self.username is None or self.password is None:
            raise ValueError("Cannot refresh_auth without username and password. Please provide these values.")
        # Login with HTTPBasic Auth to get a token:
        cred = {"username": self.username, "password": self.password}
        resp = self.request(urls.PASSWORD_LOGIN, "POST", data=cred, authenticating=True)
        assert isinstance(resp, dict)
        if resp is not None and "access_token" in resp:
            self.token = resp["access_token"]
            self.token_expiry = dateutil.parser.parse(resp["expiration"])
        else:
            raise AuthError(f"Could not authenticate: {resp}")

        # Unset BasicAuth; set Token Authorization header
        self.session.auth = None
        self.session.headers["Authorization"] = f"Bearer {self.token}"
        self._authenticated = True

    def interactive_login(self) -> Dict[str, Any]:
        """Initiate interactive login flow"""
        self.username = click.prompt("Balsam username", hide_input=False)
        self.password = click.prompt("Balsam password", hide_input=True)
        self.refresh_auth()
        click.echo("Logged In")
        return {"token": self.token, "token_expiry": self.token_expiry}


RequestsClient._client_classes["password"] = BasicAuthRequestsClient
